import os
from typing import List, Union

import h5py
import numpy as np

from ._pipeline import Pipeline
from ..tools.experiment_configuration import ExperimentConfiguration
from ..tools.experiment_iterator import GroupIterator




class Time(Pipeline):

    def run_analysis(self) -> List[np.ndarray]:
        times = []
        for g in self.experiment_configuration.experiment_iterator:
            time = self.get_time(g)
            times.append(time)
        return times


    def _validate_configuration(self):
        assert "Time" in self.experiment_configuration

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        super().__init__(experiment_configuration)
        self.output_files.update({
            "Time": os.path.join(experiment_configuration["Time"], 'time.h5')
        })

    def get_time(self, group: GroupIterator) -> np.ndarray:
        time = self._load_time(group)
        if time is not None:
            return time

        time = Time._get_group_time(group)
        group_label = group.get_label_from_values()
        self._save_time(group_label, time)
        return time

    @staticmethod
    def _get_group_time(group: GroupIterator) -> np.ndarray:
        times = []
        for sim in group:
            a = sim.analyser
            time = a.get_time()
            times.append(time)
        as_array = np.vstack(times)
        if (as_array != as_array[0]).any():
            raise RuntimeError(f"Can't get time for group '{group.values}'. "
                               "Time axes are not identical. Use custom code to read "
                               "time axes of your experiment.")
        return as_array[0]

    def _load_time(self, group: GroupIterator) -> Union[np.ndarray, None]:
        group_label = group.get_label_from_values()
        filename = self.output_files["Time"]
        if not os.path.exists(filename):
            return None
        with h5py.File(filename, 'r') as f:
            if group_label not in f:
                return None
            time = f[group_label][:]
            return time

    def _save_time(self, group_label: str, time: np.ndarray):
        filename = self.output_files["Time"]
        with h5py.File(filename, 'a') as f:
            f[group_label] = time




from abc import ABC, abstractmethod
from typing import Tuple, Dict, List
import numpy as np

from actomyosin_analyser.model.bead import Filament

_LinksArray = np.ndarray


class DataReader(ABC):
    """
    Template of a data reader class. Any derived class should
    implement the here defined abstract methods.
    This is a place holder to define the interface of a DataReader used
    by the Analyser class. Not all methods have to be overwritten, depending
    on the analyses you perform with the Analyser class. The most fundamental
    methods are marked as abstract methods (python will force you to
    implement these). Some additional methods are not implemented (raising
    exception when executed), but are not mandatory to implement (some
    analyses just won't work, when they remain not implemented).
    """

    @abstractmethod
    def __init__(self, source_file: str):
        ...

    @abstractmethod
    def get_filament_coordinates(self, minimum_image: bool) -> np.ndarray:
        pass

    def read_particle_positions(self, minimum_image: bool) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def read_time(self) -> np.ndarray:
        ...

    @abstractmethod
    def read_time_as_step_indices(self) -> np.ndarray:
        ...

    @abstractmethod
    def read_box_size(self) -> np.ndarray:
        """
        Read box size in shape (3, 2) where rows are x, y, z,
        columns are lower/upper bound.
        """
        ...

    def read_particle_types(self) -> Tuple[Dict[str, int], np.ndarray]:
        """
        Read particle types as an integer array.
        :return: Dictionary mapping particle type (str) to type as integer.
        :return: Integer array of particle types.
        """
        raise NotImplementedError

    @abstractmethod
    def get_filaments(self, frame: int) -> Tuple[List[Filament], _LinksArray]:
        ...

    @abstractmethod
    def get_filaments_all(self) -> List[List[Filament]]:
        """
        Read filament information for all frames.

        :return: Length of returned list is equal to number of recorded frames.
           For each frame, each item is a list containing Filament instances.
        """
        ...

    @abstractmethod
    def read_dt(self):
        ...

    def read_parameter(self, *args):
        raise NotImplementedError

    @abstractmethod
    def get_n_non_filament_particles(self) -> int:
        ...

    def get_motor_count(self) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def get_non_filament_coordinates(self, minimum_image):
        ...



try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata


def get_version() -> str:
    version = metadata.version('actomyosin-analyser')
    return version





from typing import Union, Tuple
import numpy as np
from scipy.optimize import curve_fit


def compute_G_via_fit_parameters(
        omega: np.ndarray,
        bead_radius: float,
        kT: float,
        A, B, C, D
) -> np.ndarray:
    """
    With parameters `A, B, C, D` retrieved via function `paust_method.fit_msd`,
    compute the predictions of complex modulus G with the Paust model.
    `omega` is `1/lag_time`.
    Make sure you provide all values in a consistent unit system (i.e. all in
    SI units, or all in some consistent internal units).

    :return: G(omega)
    """
    term1 = -kT * omega**2 * (1 + 1j * D * omega)
    term2 = np.pi * bead_radius * 1j * omega
    term3 = B + 1j * B * D * omega + 1j * omega * (A + C + 1j * A * D * omega)
    return term1 / term2 / term3


def fit_msd(
        lag_time: np.ndarray,
        msd: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Apply `curve_fit` (from `scipy.optimize`) on the Paust model for given data.

    :return: (0) The optimized parameters, and
             (1) the error covariance matrix (see documentation on `curve_fit`).
    """
    popt, err = curve_fit(paust_model, lag_time, msd,
                          sigma=lag_time,
                          maxfev=10000,
                          ftol=1e-15)
    return np.array(popt), np.array(err)


def paust_model(t: Union[float, np.ndarray],
                A: float, B: float, C: float, D: float) -> Union[float, np.ndarray]:
    """
    This function is the model for the MSD of a microrheolgy probe bead. When you use
    the function `paust_method.fit_msd`, your MSD data gets fitted to this function,
    yielding optimized values for parameters `[A, B, C, D]`.
    Optimized parameters can then be used to predict G with
    function `paust_method.compute_G_via_fit_parameters`.
    """
    return A + B * t + C * (1 - np.exp(-t / D))



import numpy as np


class MayaviAnimator:

    def __init__(self):
        self.data_reader = None

    def show_at_t(self, t):
        raise NotImplementedError("Abstract base class")

    def _data_slice_to_coords(self, data_slice: np.ndarray):
        size = data_slice.shape[0] // 3
        x = np.empty(size)
        y = np.empty(size)
        z = np.empty(size)        

        for i in range(size):
            x[i] = data_slice[3 * i]
            y[i] = data_slice[3 * i + 1]
            z[i] = data_slice[3 * i + 2]
        return x, y, z



import json


class ConfigReader:

    def __init__(self, fname):

        f = open(fname, "rt")
        self.json = json.load(f)

    def get_number_of_time_steps(self):        
        return self.json["n_time_steps"]
        

    def get_force_constants_actin(self):
        bend = self.json["force_constants"]["actin"]["bend"]               
        random_force_factor = self.json["force_constants"]["actin"]["random_force_factor"]
        resting_length = self.json["force_constants"]["actin"]["resting_length"]     
        stretch = self.json["force_constants"]["actin"]["stretch"]
        return (bend, random_force_factor, resting_length, stretch)

    def get_force_constants_van_der_waals(self):
        cut_off = self.json["force_constants"]["van_der_waals"]["cut_off"]
        distance_factor = self.json["force_constants"]["van_der_waals"]["distance_factor"]
        return (cut_off, distance_factor)

    def get_actin_properties(self):
        dt_by_drag_coefficient = self.json["actin_properties"]["dt_by_drag_coefficient"]     
        return dt_by_drag_coefficient

    def get_box_sizes(self):
        size_x = self.json["box"]["size_x"]
        size_y = self.json["box"]["size_y"]
        size_z = self.json["box"]["size_z"]
        return(size_x, size_y, size_z)



import os
from multiprocessing import Pool
from typing import Sequence

import numpy as np
from actomyosin_analyser.tools.experiment_iterator import GroupIterator, Simulation

from actomyosin_analyser.tools.group_results_container import GroupResultsContainer

from actomyosin_analyser.tools.experiment_configuration import ExperimentConfiguration

from actomyosin_analyser.pipelines._pipeline import Pipeline

EndToEndContainer = GroupResultsContainer[np.ndarray]


class EndToEndDistance(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration, n_processes: int=1):
        super().__init__(experiment_configuration)
        self._blob_folder = os.path.join(experiment_configuration['EndToEndDistance'], 'blobs')
        self._n_processes = n_processes

    def run_analysis(self, frames: Sequence[int], n_beads: int) -> EndToEndContainer:
        return self.get_end_to_end_distances(frames, n_beads)

    def _validate_configuration(self):
        assert "EndToEndDistance" in self.experiment_configuration

    def get_end_to_end_distances(
            self,
            frames: Sequence[int],
            n_beads: int
    ) -> EndToEndContainer:
        container = EndToEndContainer()

        for group in self.experiment_configuration.experiment_iterator.groups:
            e2e = self._get_end_to_end_distances_of_group(group, frames, n_beads)
            container[group] = e2e

        return container

    def _get_end_to_end_distances_of_group(
            self,
            group: GroupIterator,
            frames: Sequence[int],
            n_beads: int
    ) -> np.ndarray:
        simulations = [sim for sim in group]

        caller = _E2ECaller(frames, n_beads)

        pool = Pool(self._n_processes)
        results = pool.map(
            caller,
            simulations
        )

        e2e_distances = np.concatenate(results)
        return e2e_distances


class _E2ECaller:

    def __init__(self, frames: Sequence[int], n_beads: int):
        self.frames = frames
        self.n_beads = n_beads

    def __call__(self, simulation: Simulation) -> np.ndarray:
        e2e_distances = []
        a = simulation.analyser

        for f in self.frames:
            e2e_f = a.get_end_to_end_distances(f, self.n_beads)
            e2e_distances.append(e2e_f)

        return np.concatenate(e2e_distances)



import os
from abc import ABC, abstractmethod
from typing import Callable, Sequence, Any, Dict, List

import numpy as np
from actomyosin_analyser.model.bead import Filament
from scipy.stats import gaussian_kde


class AbstractAnalyser(ABC):

    def __init__(self, data_root: str):
        self._data_root = data_root
        if not os.path.isdir(self._data_root):
            raise FileNotFoundError(f"{self._data_root} folder does not exist")
        self.blob_root = os.path.join(self._data_root, 'analysis_blobs')
        os.makedirs(self.blob_root, exist_ok=True)

    def _get_blob_analysis_dataset(
            self,
            dset_name: str,
            compute_function: Callable[[], np.ndarray],
    ) -> np.ndarray:
        blob_file = os.path.join(self.blob_root, dset_name + '.npy')
        if os.path.exists(blob_file):
            return np.load(blob_file)
        data = compute_function()
        np.save(blob_file, data)
        return data

    def _get_nested_blob_analysis_dataset(
            self,
            dset_name: str,
            compute_function: Callable[[Any], np.ndarray],
            keyword_args: Dict[str, int]
    ):
        blob_folder = os.path.join(self.blob_root, dset_name)
        file_name_from_args = '_'.join([f'{key}_{keyword_args[key]}'
                                        for key in sorted(keyword_args.keys())])
        blob_file = os.path.join(blob_folder, file_name_from_args + '.npy')
        if os.path.exists(blob_file):
            return np.load(blob_file)
        data = compute_function(**keyword_args)
        os.makedirs(blob_folder, exist_ok=True)
        np.save(blob_file, data)
        return data

    @abstractmethod
    def get_links(self, frame: int) -> np.ndarray:
        ...

    @abstractmethod
    def get_trajectories_non_filament(self, minimum_image: bool) -> np.ndarray:
        ...

    @abstractmethod
    def get_positions_of_filaments_at_frames(
            self,
            frames: Sequence[int],
            minimum_image: bool
    ) -> np.ndarray:
        ...

    @abstractmethod
    def get_densities_gaussian_kde_scipy(
            self,
            frame: int,
            bandwidth: float
    ) -> gaussian_kde:
        ...

    @abstractmethod
    def get_filaments(self) -> List[List[Filament]]:
        ...

    @abstractmethod
    def get_trajectories_filaments(self, minimum_image: bool) -> np.ndarray:
        """
        Get trajectories of filaments. Filaments are represented by multiple coordinates,
        which corresponds to coordinates of beads in bead chains (e.g. from simulations
        with the ``bead_state_model`` framework)
        , or start and end points
        of segments in segment representations of filaments (e.g. from simulations with
        the ``cytosim`` framework).
        Which coordinates belong to which filament can be figured out with the return
        value of the ``get_filaments`` method.

        :param minimum_image: Set ``True`` to project coordinates that exceed
           the simulation box back into the simulation box.
        :return: Numpy array with 3 dimensions:
           (number of frames, number of filament particles, xyz-coordinates).
        """
        ...

    @abstractmethod
    def read_simulation_box(self) -> np.ndarray:
        ...



from typing import List, Tuple, Union
import os

import h5py
import numpy as np
import matplotlib.pyplot as plt
from ._pipeline import Pipeline
from .time import Time
from ..tools.experiment_configuration import ExperimentConfiguration
from ..tools.experiment_iterator import GroupIterator
from ..analysis.analyser import Analyser


class Energy(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        super().__init__(experiment_configuration)
        self.plot_files.update({
            "energy": os.path.join(experiment_configuration["Energy"], 'energy.svg'),
            "individual_energies": os.path.join(experiment_configuration["Energy"],
                                                "energies_{label}.svg")
        })
        self.output_files.update({
            "energy": os.path.join(experiment_configuration['Energy'], 'energy.h5')
        })

    def run_analysis(self):
        times, energies = self.get_energies()
        self._plot_energies(times, energies)

    def _validate_configuration(self):
        assert "Energy" in self.experiment_configuration

    def get_mean_energies(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        times, mean_energies = [], []
        for group in self.experiment_configuration.experiment_iterator.groups:
            t, me = self._get_time_and_mean_energies_group(group)
            times.append(t)
            mean_energies.append(me)
        return times, mean_energies

    def _get_time_and_mean_energies_group(
            self,
            group: GroupIterator
    ) -> Tuple[np.ndarray, np.ndarray]:
        time, mean_e = self._load_time_and_mean_energies(group.get_label_from_values())
        if time is not None:
            return time, mean_e
        time = Time(self.experiment_configuration).get_time(group)
        energies = Energy._get_group_energy(group)
        mean_energies = energies.mean(0)
        self._save_time_and_mean_energies(
            group.get_label_from_values(),
            time, mean_energies
        )
        return time, mean_energies

    def get_energies(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        times = []
        energies = []
        iterator = self.experiment_configuration.experiment_iterator
        for group in iterator:
            time = Time(self.experiment_configuration).get_time(group)
            times.append(time)
            e_group = Energy._get_group_energy(group)
            energies.append(e_group)
        return times, energies

    @staticmethod
    def _get_group_energy(group: GroupIterator) -> np.ndarray:
        energy = None
        for i, sim in enumerate(group):
            a = sim.analyser  # type: Analyser
            e_i = a.get_energies_stretch_and_bend()
            k_bend = a.data_reader.read_parameter('k_bend')
            k_stretch = a.data_reader.read_parameter('k_stretch')
            e_i[:, 0] = e_i[:, 0] * k_stretch
            e_i[:, 1] = e_i[:, 1] * k_bend
            if energy is None:
                shape = [len(group)] + list(e_i.shape)
                energy = np.full(shape, np.nan)
            energy[i] = e_i
        return energy

    def _plot_energies(self, times: List[np.ndarray], energies: List[np.ndarray]):
        groups = self.experiment_configuration.experiment_iterator.groups
        fig, ax = plt.subplots(1, 1)
        label_stretch = '$E_\\mathrm{stretch}$, '
        label_bend = '$E_\\mathrm{bend}$, '
        for i in range(len(groups)):
            group = groups[i]
            color = group.color
            if color is None:
                color = f'C{i}'
            label = group.label
            e = energies[i]
            t = times[i]
            self._plot_individual_energies(e, group)
            mean_e = e.mean(0)
            ax.plot(t, mean_e[:, 0], '--', color=color, label=label_stretch)
            ax.plot(t, mean_e[:, 1], '-', color=color, label=label_bend + label)
            label_stretch = None
            label_bend = ''
        ax.legend()
        ax.set(
            xlabel='$t/t_0$',
            ylabel='$\\langle E \\rangle$ /kT',
            title="Energy"
        )
        fig.tight_layout()
        fig.savefig(self.plot_files['energy'])



    def _plot_individual_energies(self, energies: np.ndarray, group: GroupIterator):
        fig, ax = plt.subplots(1, 1)
        label = group.label
        file_label = group.get_label_from_values()
        label_stretch = '$E_\\mathrm{stretch}$'
        label_bend = '$E_\\mathrm{bend}$'
        for i, e_i in enumerate(energies):
            color = f'C{i}'
            ax.plot(e_i[:, 0], '--', color=color, label=label_stretch)
            ax.plot(e_i[:, 1], '-', color=color, label=label_bend)
            label_stretch = None
            label_bend = None
        ax.set(
            xlabel='frame',
            ylabel='$E$ / kT',
            title=label
        )
        ax.legend()
        fig.tight_layout()
        filename = self.plot_files['individual_energies'].format(label=file_label)
        fig.savefig(filename)

    def _load_time_and_mean_energies(
            self,
            group_label: str
    ) -> Union[Tuple[None, None], Tuple[List[np.ndarray], List[np.ndarray]]]:
        if not os.path.exists(self.output_files['energy']):
            return None, None
        with h5py.File(self.output_files['energy'], 'r') as file:
            if 'mean' not in file:
                return None, None
            mean_h5 = file['mean']
            if group_label not in mean_h5:
                return None, None
            times = mean_h5[group_label]['time'][:]
            mean_energies = mean_h5[group_label]['mean_energy'][:]
        return times, mean_energies

    def _save_time_and_mean_energies(
            self,
            group_label: str,
            times: np.ndarray,
            mean_energies: np.ndarray
    ):
        with h5py.File(self.output_files['energy'], 'w') as file:
            mean_h5 = file.create_group('mean')
            g_h5 = mean_h5.create_group(group_label)
            g_h5['time'] = times
            g_h5['mean_energy'] = mean_energies



from .vtk_filament_representation import VTKFilamentRepresentation


class FilamentAnimator:

    def __init__(self):
        self.filament_data_reader = None
        self.filaments = []
        self.vtk_renderer = None

    def show_at_t(self, t):
        data_slice, filament_length_array = self.filament_data_reader\
                                                .get_coordinates_at_time_step(t)
        self._draw_data_slice(data_slice, filament_length_array)

    def _draw_data_slice(self, data_slice, filament_length_array):
        if data_slice.shape[0] % 3 != 0:
            msg = "data_slice needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)
        if len(filament_length_array) > len(self.filaments):
            for i in range(len(self.filaments), len(filament_length_array)):
                new_filament = VTKFilamentRepresentation()
                new_filament.vtk_renderer = self.vtk_renderer
                self.filaments.append(new_filament)

        current_index = 0
        i = 0
        for f in filament_length_array:
            print(current_index, i, f)
            self.filaments[current_index].set_coordinates(data_slice[i: i + 3 * f])
            i += 3 * f
            current_index += 1



import numpy as np


class TextStackReader:
    def __init__(self, folder):
        self.beads_reader = _BeadsReader(folder + "/beads.txt")
        self.beads_reader.read()
        self.filaments_reader = _FilamentsReader(folder + "/filaments.txt")
        self.filaments_reader.read()


class _BaseReader:

    def __init__(self, fname):
        self.fname = fname
        self.t = None

    def read(self):
        f = open(self.fname, "rt")
        lines = f.readlines()

        t = []
        N = []

        i = 0
        while i < len(lines):
            t_i, N_i = self.parse_step_line(lines[i])
            t.append(t_i)
            N.append(N_i)
            self.parse_content_following_step_line(t_i, lines[i + 1: i + N_i + 1])
            i += N_i + 1
        self.t = np.array(t)

    def parse_content_following_step_line(self, t, content_lines):
        raise NotImplementedError("Abstract Base Class")

    @staticmethod
    def parse_step_line(step_line):
        split = step_line.split("\t")
        t = int(split[0].split("=")[1])
        N = int(split[1].split("=")[1])
        return t, N


class _FilamentsReader(_BaseReader):

    def __init__(self, fname):
        _BaseReader.__init__(self, fname)
        self.filaments = {}

    def parse_content_following_step_line(self, t, lines):
        N = len(lines)
        ids = np.empty(N, dtype=int)
        firsts = np.empty(N, dtype=np.int32)
        lasts = np.empty(N, dtype=np.int32)
        counts = np.empty(N, dtype=np.int32)
        for i, l in enumerate(lines):
            split = l.split("\t")
            ids[i] = int(split[0])
            firsts[i] = int(split[1])
            lasts[i] = int(split[2])
            counts[i] = int(split[3])
        self.filaments[t] = _Filaments(ids, firsts, lasts, counts)


class _BeadsReader(_BaseReader):

    def __init__(self, fname):
        _BaseReader.__init__(self, fname)
        self.beads = {}

    def parse_content_following_step_line(self, t, lines):
        N = len(lines)
        ids = np.empty(N, dtype=int)
        x = np.empty(N)
        y = np.empty(N)
        z = np.empty(N)
        links_prev = np.empty(N, dtype=np.int32)
        links_next = np.empty(N, dtype=np.int32)
        links_crossf = np.empty(N, dtype=np.int32)
        for i, l in enumerate(lines):
            split = l.split("\t")
            ids[i] = int(split[0])
            x[i] = float(split[1])
            y[i] = float(split[2])
            z[i] = float(split[3])
            links_prev[i] = int(split[4])
            links_next[i] = int(split[5])
            links_crossf[i] = int(split[6])
        b = _Beads(ids, x, y, z, links_prev, links_next, links_crossf)
        self.beads[t] = b


class _Filaments:
    def __init__(self, ids: np.ndarray,
                 firsts: np.ndarray, lasts: np.ndarray,
                 counts: np.ndarray):
        self.ids = ids
        self.firsts = firsts
        self.lasts = lasts
        self.counts = counts

    def __str__(self):
        s = "Data for {} filaments."
        return s.format(len(self.ids))

    def __repr__(self):
        return self.__str__()


class _Beads:

    def __init__(self, ids: np.ndarray,
                 x: np.ndarray, y: np.ndarray, z: np.ndarray,
                 links_prev: np.ndarray, links_next: np.ndarray,
                 links_crossf: np.ndarray):
        self.ids = ids
        self.x = x
        self.y = y
        self.z = z
        self.links_prev = links_prev
        self.links_next = links_next
        self.links_crossf = links_crossf

    def __str__(self):
        s = "Actins data consisting of {} beads."
        return s.format(len(self.x))

    def __repr__(self):
        return self.__str__()



from abc import ABC, abstractmethod
import os

from ..tools.experiment_configuration import ExperimentConfiguration


class Pipeline(ABC):

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        self.experiment_configuration = experiment_configuration
        self._validate_configuration()
        self.plot_files = {}
        self.input_files = {}
        self.output_files = {}

    def __call__(self, *args, **kwargs):
        return self.run_analysis(*args, **kwargs)

    @abstractmethod
    def run_analysis(self, *args, **kwargs):
        ...

    @abstractmethod
    def _validate_configuration(self):
        ...

    def _remove_output_files(self):
        for key in self.output_files:
            f = self.output_files[key]
            if os.path.exists(f) and os.path.isfile(f):
                os.remove(f)


from typing import Union, List, Tuple
import os
from multiprocessing.pool import ThreadPool
import logging

import h5py
import numpy as np

from actomyosin_analyser.pipelines._pipeline import Pipeline
from actomyosin_analyser.pipelines.uniform_filaments.filament_trajectories import FilamentTrajectories
from actomyosin_analyser.tools.experiment_configuration import ExperimentConfiguration
from actomyosin_analyser.tools.experiment_iterator import GroupIterator

logger = logging.getLogger(__name__)


class FilamentMSDs(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration, overwrite: bool=False):
        super().__init__(experiment_configuration)
        self._trajectories = FilamentTrajectories(experiment_configuration)
        self.output_files.update({
            'SelectedMSDs': os.path.join(experiment_configuration['FilamentMSDs'],
                                         'selected_msds.h5')
        })
        self.skip = experiment_configuration.skip_n_frames

        if overwrite:
            self._remove_output_files()

    def run_analysis(self, fixed_filament_length: int, use_beads_of_filament: Union[List[int], np.ndarray],
                     number_of_parallel_processes: int = 1) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
        return self.get_msds(fixed_filament_length, use_beads_of_filament, number_of_parallel_processes)

    def _validate_configuration(self):
        assert 'FilamentMSDs' in self.experiment_configuration

    def get_msds(self, fixed_filament_length: int, use_beads_of_filament: Union[List[int], np.ndarray],
                 number_of_parallel_processes: int = 1) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
        if number_of_parallel_processes > 1:
            return self._get_msds_parallel(number_of_parallel_processes,
                                           fixed_filament_length, use_beads_of_filament)
        lag_times, msds, variances = [], [], []
        for group in self.experiment_configuration.experiment_iterator:
            lag_time, msd, var_msd = self.get_msds_of_single_group(group, fixed_filament_length,
                                                                   use_beads_of_filament)
            lag_times.append(lag_time)
            msds.append(msd)
            variances.append(var_msd)
        return lag_times, msds, variances

    def get_msds_of_single_group(
            self, group: GroupIterator,
            fixed_filament_length: int,
            use_beads_of_filament: Union[List[int], np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        h5_group_name = FilamentTrajectories.create_label_from_parameters(
            fixed_filament_length, use_beads_of_filament
        )
        lag_time, msd, var_msd = self._load_msds(group, h5_group_name)
        if lag_time is not None:
            return lag_time, msd, var_msd
        time, trajectory = self._trajectories.get_trajectories_of_single_group(
            group, fixed_filament_length, use_beads_of_filament
        )
        if self.skip == 0:
            lag_time = time[1:]
        else:
            lag_time = time[1:-self.skip]
        trajectory = trajectory[:, self.skip:]
        msd, var_msd = self._compute_msds_of_single_group(trajectory)
        self._save_msds_of_single_group(group, h5_group_name, lag_time, msd, var_msd)
        return lag_time, msd, var_msd

    @staticmethod
    def _compute_msds_of_single_group(
            trajectory: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        _number_of_simulations = trajectory.shape[0]
        _number_of_frames = trajectory.shape[1]
        shape = (_number_of_simulations,
                 _number_of_frames - 1)
        msd = np.empty(shape)
        var_msd = np.empty(shape)
        for lag in range(1, _number_of_frames):
            diff = trajectory[:, lag:] - trajectory[:, :-lag]
            displacements = (diff ** 2).sum(3)
            d_reshaped = displacements.reshape((displacements.shape[0], -1))
            msd[:, lag - 1] = np.mean(d_reshaped, axis=1)
            var_msd[:, lag - 1] = np.var(d_reshaped, axis=1)
        return msd, var_msd

    def _load_msds(
            self, group: GroupIterator,
            h5_group_name: str
    ) -> Union[Tuple[np.ndarray, np.ndarray, np.ndarray],
               Tuple[None, None, None]]:
        fname = self.output_files["SelectedMSDs"]
        if not os.path.exists(fname):
            return None, None, None
        with h5py.File(fname, 'r') as file:
            group_skip_name = f'frames_skipped_{self.skip}'
            if group_skip_name not in file:
                return None, None, None
            group_skip = file[group_skip_name]
            if h5_group_name not in group_skip:
                return None, None, None
            h5_group = group_skip[h5_group_name]
            label = group.get_label_from_values()
            if label not in h5_group:
                return None, None, None
            msd = h5_group[label]['msd'][:]
            lag_time = h5_group[label]['lag_time'][:]
            var_msd = h5_group[label]['var_msd'][:]
        return lag_time, msd,  var_msd

    def _save_msds_of_single_group(
            self, group: GroupIterator,
            h5_group_name: str,
            lag_time: np.ndarray,
            msd: np.ndarray,
            var_msd: np.ndarray
    ):
        with h5py.File(self.output_files["SelectedMSDs"], 'a') as file:
            name_group_skip = f'frames_skipped_{self.skip}'
            if name_group_skip in file:
                group_skip = file[name_group_skip]
            else:
                group_skip = file.create_group(name_group_skip)
            if h5_group_name in group_skip:
                h5_group = group_skip[h5_group_name]
            else:
                h5_group = group_skip.create_group(h5_group_name)
            label = group.get_label_from_values()
            if label in h5_group:
                raise RuntimeError("msd of group seem to be already present "
                                   f"in {self.output_files['SelectedTrajectories']}"
                                   f":{h5_group_name}/{label}")
            h5_gg = h5_group.create_group(label)
            h5_gg['msd'] = msd
            h5_gg['lag_time'] = lag_time
            h5_gg['var_msd'] = var_msd

    def _get_msds_parallel(
            self,
            number_of_parallel_processes,
            fixed_filament_length,
            use_beads_of_filament
    ) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
        lag_times, msds, variances = [], [], []
        groups = self.experiment_configuration.experiment_iterator
        h5_group_name = FilamentTrajectories.create_label_from_parameters(
            fixed_filament_length, use_beads_of_filament
        )
        for group in groups:
            t, msd, var_msd = self._load_msds(group, h5_group_name)
            lag_times.append(t)
            msds.append(msd)
            variances.append(var_msd)

        groups_not_loaded = []
        indices_not_loaded = []
        for i, group in enumerate(groups):
            if lag_times[i] is None:
                groups_not_loaded.append(group)
                indices_not_loaded.append(i)

        lag_times_compute, trajectories = [], []
        for group in groups_not_loaded:
            t, traj = self._trajectories.get_trajectories_of_single_group(
                group,
                fixed_filament_length,
                use_beads_of_filament
            )
            t = t[1:-self.skip]
            traj = traj[:, self.skip:]
            lag_times_compute.append(t)
            trajectories.append(traj)

        pool = ThreadPool(number_of_parallel_processes)
        results = pool.map(FilamentMSDs._compute_msds_of_single_group, trajectories)

        for i, idx in enumerate(indices_not_loaded):
            lag = lag_times_compute[i]
            msd, var_msd = results[i]
            group = groups_not_loaded[i]
            self._save_msds_of_single_group(group, h5_group_name,
                                            lag, msd, var_msd)
            lag_times[idx] = lag
            msds[idx] = msd
            variances[idx] = var_msd
        return lag_times, msds, variances



import json
import numpy as np


class ConfigReaderBeads:

    param_array_dtype = [('assembly_rate', float),
                         ('box_x', float), ('box_y', float), ('box_z', float),
                         ('disassembly_rate', float),
                         ('k_bend', float), ('k_stretch', float),
                         ('n_beads', int),
                         ('n_steps', int),
                         ('nucleation_rate', float),
                         ('time_step', float)]

    old_param_array_dtype = [('assembly_rate', float),
                             ('box_x', float), ('box_y', float), ('box_z', float),
                             ('disassembly_rate', float),
                             ('k_bend', float), ('k_stretch', float),
                             ('n_beads', int),
                             ('n_steps', int),
                             ('nucleation_rate', float),
                             ('random_step_size', float),
                             ('segment_length', float)]
    
    def __init__(self, fname):        
        with open(fname, "rt") as f:
            self.json = json.load(f)

    def get_number_of_time_steps(self):        
        return self.json["n_steps"]

    def get_box_size(self) -> np.ndarray:
        size_x = self.json["box_size"]["x"]
        size_y = self.json["box_size"]["y"]
        size_z = self.json["box_size"]["z"]
        return np.array([size_x, size_y, size_z])

    def get_bead_diameter(self):
        if 'bead_diameter' in self.json:
            return self.json['bead_diameter']
        return 1.0

    def get_time_step(self):
        if 'time_step' in self.json:
            return self.json['time_step']
        return 1.0

    def get_temperature(self):
        return self.json['temperature']
    
    def get_random_step_size(self):
        return self.json['random_step_size']

    def get_k_stretch(self):
        return self.json['k_stretch']

    def get_k_bend(self):
        return self.json['k_bend']

    def get_segment_length(self):
        return self.json['segment_length']

    def config_to_param_array(self, old_dtype=False) -> np.ndarray:
        if not old_dtype:
            return np.array((self.json['assembly_rate'],
                             self.json['box_size']['x'],
                             self.json['box_size']['y'],
                             self.json['box_size']['z'],
                             self.json['disassembly_rate'],
                             self.json['k_bend'],
                             self.json['k_stretch'],
                             self.json['n_beads'],
                             self.json['n_steps'],
                             self.json['nucleation_rate'],
                             self.json['time_step']),
                            dtype=self.param_array_dtype)
        return np.array((self.json['assembly_rate'],
                         self.json['box_size']['x'],
                         self.json['box_size']['y'],
                         self.json['box_size']['z'],
                         self.json['disassembly_rate'],
                         self.json['k_bend'],
                         self.json['k_stretch'],
                         self.json['n_beads'],
                         self.json['n_steps'],
                         self.json['nucleation_rate'],
                         self.json['random_step_size'],
                         self.json['segment_length']),
                        dtype=self.old_param_array_dtype)



import os
from typing import Union, List, Tuple, Dict, Any

import h5py
import numpy as np
import matplotlib.pyplot as plt

from actomyosin_analyser.pipelines._pipeline import Pipeline
from actomyosin_analyser.pipelines.time import Time
from actomyosin_analyser.tools.experiment_configuration import ExperimentConfiguration
from actomyosin_analyser.tools.experiment_iterator import GroupIterator
from actomyosin_analyser.analysis.analyser import Analyser




class FilamentTrajectories(Pipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration):
        super().__init__(experiment_configuration)
        self.output_files.update({
            'SelectedTrajectories': os.path.join(experiment_configuration['FilamentTrajectories'],
                                                 'selected_trajectories.h5')
        })
        self.plot_files.update({
            'RandomSetOfTrajectories': os.path.join(experiment_configuration['FilamentTrajectories'],
                                                    'random_trajectories_{label}.svg')
        })

    def _validate_configuration(self):
        assert "FilamentTrajectories" in self.experiment_configuration

    def run_analysis(self, fixed_filament_length: int,
                     use_beads_of_filament: Union[List[int], np.ndarray]) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        return self.get_trajectories(fixed_filament_length, use_beads_of_filament)

    def get_trajectories(
            self, fixed_filament_length: int,
            use_beads_of_filament: Union[List[int], np.ndarray]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        groups = self.experiment_configuration.experiment_iterator.groups
        times, trajectories = [], []
        for g in groups:
            time, traj = self.get_trajectories_of_single_group(
                g, fixed_filament_length,
                use_beads_of_filament
            )
            times.append(time)
            trajectories.append(traj)
        return times, trajectories

    def get_trajectories_of_single_group(
            self, group,
            fixed_filament_length,
            use_beads_of_filament
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get trajectories of filaments of a group of simulations. The beads of the
        filaments are selected via `use_beads_of_filaments`. The beads per filament
        has to be fixed, this number has to be specified via parameter `fixed_filament_length`.

        :return: (0) time. 1D array of time stamps.
                 (1) trajectories. 4D array of filament bead trajectories.
                     Dimensions: (0) Number of simulations in group,
                                 (1) number of recorded frames,
                                 (2) number of particles (len(use_beads_of_filament) * n_filaments)
                                 (3) 2 or 3 spatial coordinates (x, y, z)
        """
        time, traj = self._load_trajectories(group, fixed_filament_length, use_beads_of_filament)
        if traj is not None:
            return time, traj
        time = Time(self.experiment_configuration).get_time(group)
        traj = self._read_trajectories(group, fixed_filament_length, use_beads_of_filament)
        self._save_trajectories(time, traj, group, fixed_filament_length, use_beads_of_filament)
        return time, traj

    def plot_random_set_of_trajectories(self, trajectories: List[np.ndarray], n_trajectories: int,
                                        plot_kwargs: Dict[str, Any]={}):
        groups = self.experiment_configuration.experiment_iterator.groups
        for i in range(len(groups)):
            g = groups[i]
            traj = trajectories[i]
            self._plot_random_set_of_trajectories_group(traj, g, n_trajectories, plot_kwargs)

    def _plot_random_set_of_trajectories_group(self, trajectories: np.ndarray,
                                               group: GroupIterator, n_trajectories: int,
                                               plot_kwargs: Dict[str, Any]={}):
        _number_of_simulations = trajectories.shape[0]
        _number_of_particles = trajectories.shape[2]
        sim = np.random.randint(0, _number_of_simulations, n_trajectories)
        particle = np.random.randint(0, _number_of_particles, n_trajectories)
        selected_trajectories = trajectories[sim, :, particle]
        offset = selected_trajectories[:, 0, :]
        selected_trajectories = selected_trajectories - offset[:, None, :]

        fig, ax = plt.subplots(1, 1)

        for traj in selected_trajectories:
            ax.plot(traj[:, 0], traj[:, 1], **plot_kwargs)

        ax.set(
            aspect='equal',
            xlabel='x / x0',
            ylabel='y / x0',
            title=group.label
        )
        label = group.get_label_from_values()
        fname = self.plot_files['RandomSetOfTrajectories'].format(label=label)
        fig.tight_layout()
        fig.savefig(fname)

    @staticmethod
    def _read_trajectories(group: GroupIterator,
                           fixed_filament_length: int,
                           use_beads_of_filament: Union[List[int], np.ndarray]):
        trajectories = None
        mask = None
        for i, sim in enumerate(group):
            a = sim.analyser  # type: Analyser
            traj_i = a.get_trajectories_filaments(minimum_image=False)
            if mask is None:
                assert traj_i.shape[1] % fixed_filament_length == 0
                mask = []
                for b in use_beads_of_filament:
                    mask += list(range(b, traj_i.shape[1], fixed_filament_length))
                mask = np.array(mask)
            if trajectories is None:
                trajectories = np.full((len(group), traj_i.shape[0], len(mask), traj_i.shape[2]), np.nan)
            trajectories[i] = traj_i[:, mask]
        return trajectories

    def _load_trajectories(
            self, group: GroupIterator, fixed_filament_length: int,
            use_beads_of_filament: Union[List[int], np.ndarray]
    ) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[None, None]]:
        fname = self.output_files['SelectedTrajectories']
        if not os.path.exists(fname):
            return None, None
        with h5py.File(fname, 'r') as file:
            h5_group_name = FilamentTrajectories.create_label_from_parameters(
                fixed_filament_length,
                use_beads_of_filament
            )
            if h5_group_name not in file:
                return None, None
            h5_group = file[h5_group_name]
            label = group.get_label_from_values()
            if label not in h5_group:
                return None, None
            traj = h5_group[label]['trajectories'][:]
            time = h5_group[label]['time'][:]
        return time, traj

    def _save_trajectories(self, time: np.ndarray,
                           trajectories: np.ndarray,
                           group: GroupIterator, fixed_filament_length: int,
                           use_beads_of_filament: Union[List[int], np.ndarray]):
        with h5py.File(self.output_files['SelectedTrajectories'], 'a') as file:
            h5_group_name = FilamentTrajectories.create_label_from_parameters(fixed_filament_length,
                                                                              use_beads_of_filament)
            if h5_group_name in file:
                h5_group = file[h5_group_name]
            else:
                h5_group = file.create_group(h5_group_name)

            label = group.get_label_from_values()
            if label in h5_group:
                raise RuntimeError("trajectories of group seem to be already present "
                                   f"in {self.output_files['SelectedTrajectories']}"
                                   f":{h5_group_name}/{label}")
            h5_gg = h5_group.create_group(label)
            h5_gg['trajectories'] = trajectories
            h5_gg['time'] = time

    @staticmethod
    def create_label_from_parameters(fixed_filament_length: int,
                                     use_beads_of_filament: Union[List[int], np.ndarray]) -> str:
        if not isinstance(fixed_filament_length, int):
            raise TypeError("parameter fixed_filament_length has to be an integer")
        label = str(fixed_filament_length)
        for item in use_beads_of_filament:
            if not isinstance(item, int):
                raise TypeError("all items of list/array use_beads_of_filament have to be integers")
            label += '_' + str(item)
        return label



# coding: utf-8
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.widgets import Slider


def get_coordinates_slice_as_plottable_arrays(coordinates_slice):
    if len(coordinates_slice) % 3 != 0:
        raise RuntimeError("number of items in array not dividable by 3!")
    n_points = int(len(coordinates_slice) / 3)
    x = []
    y = []
    z = []
    for i in range(0, n_points * 3, 3):
        x.append(coordinates_slice[i])
        y.append(coordinates_slice[i + 1])
        z.append(coordinates_slice[i + 2])
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)    
    return x, y, z


def plot_segment(ax, slice_of_filament_array: np.ndarray):
    x, y, z = get_coordinates_slice_as_plottable_arrays(slice_of_filament_array)
    line = ax.plot(x, y, z, marker="o")
    return line[0]


class PlotDings:

    def __init__(self, time_steps: int):
        self.time_steps = time_steps
        self.f = plt.figure()
        self.ax = self.f.gca(projection='3d')
        self.f.subplots_adjust(left=0.03, bottom=0.2)

        self.actin_data_reader = None
        self.segment_lines = []

        axcolor = 'lightgoldenrodyellow'
        self.ax_time = self.f.add_axes([0.25, 0.1, 0.65, 0.03],
                                       facecolor=axcolor)
        self.slider_time = Slider(self.ax_time, 
                                  r"time [$\Delta t$]",
                                  0, time_steps - 1, valinit=0,
                                  valfmt='%0.0f')
        self.slider_time.on_changed(self._update)

    def _update(self, val):        
        i = 0
        t = int(self.slider_time.val)
        data_slice = self.actin_data_reader.get_coordinates_at_time_step(t)
        for i in range(len(self.segment_lines)):
            l = self.segment_lines[i]
            j = i * 6
            new_coordinates = get_coordinates_slice_as_plottable_arrays(data_slice[j: j + 6])
            # l._offsets3d = new_coordinates
            l.set_data(new_coordinates[0], new_coordinates[1])
            l.set_3d_properties(new_coordinates[2])
        self.f.canvas.draw_idle()

    def _initialize_plot(self):
        data_slice = self.actin_data_reader.get_coordinates_at_time_step(0)
        for i in range(0, data_slice.shape[0], 6):
            l = plot_segment(self.ax, data_slice[i: i + 6])
            self.segment_lines.append(l)

    def show(self):
        self._initialize_plot()
        self.f.show()



from typing import Union
import numpy as np
from scipy.integrate import simps

a = 14.054
b = 0.473
cij = np.array([[-0.75, 0.359375, -0.109375],
                [-0.5, 1.0625, -0.5625]])


def _c(k: float) -> float:
    return 1 - (1 + (0.38 * k ** -0.95) ** -5) ** -0.2


def _d(k: float) -> float:
    if k < 0.125:
        return 0.0
    return 1 - 1 / (0.177 / (k - 0.111) + 6.4 * (k - 0.111) ** 0.783)


def _sum1(i: int, k: float, r: float) -> float:
    return cij[i + 1, 0] * k ** i * r ** 2 + cij[i + 1, 1] * k ** i * r ** 4 + cij[i + 1, 2] * k ** i * r ** 6


def _J_SYD(k: float) -> float:
    if k <= 0.125:
        return (3 / 4 / np.pi / k) ** 1.5 * np.exp(0) * (1 - 1.25 * k)
    else:
        return 112.04 * k ** 2 * np.exp(0.246 / k - a * k)


def _QI(r: Union[float, np.ndarray], k: float) -> Union[float, np.ndarray]:
    return _J_SYD(k) * ((1 - _c(k) * r ** 2) / (1 - r ** 2)) ** 2.5 * np.exp(
        (_sum1(-1, k, r) + _sum1(0, k, r)) / (1 - r ** 2)) \
           * np.exp(-(_d(k) * k * a * b * (1 + b) * r ** 2) / (1 - b ** 2 * r ** 2)) * np.i0(
        -(_d(k) * k * a * (1 + b) * r) / (1 - b ** 2 * r ** 2))


def CDF_QI(r: np.ndarray, k: float) -> np.ndarray:
    dQI = []
    for i in range(1, len(r)):
        dQIi = simps(_QI(r[:i], k) * r[:i] ** 2, r[:i])
        dQI.append(dQIi)
    CDF = np.array(dQI)
    return CDF / CDF[-1]


def mean_relative_end_to_end_distance(k: float, num_points: int = 500) -> float:
    r = np.linspace(0, 1, num_points)
    qi = _QI(r, k)
    if np.isnan(qi[-1]):
        qi = qi[:-1]
        r = r[:-1]
    R = 4 * np.pi * simps(qi * r ** 3, r)
    return R



import os.path

import numpy as np

from actomyosin_analyser.analysis.abstract_analyser import AbstractAnalyser


class StateExporter:

    def __init__(self, analyser: AbstractAnalyser):
        self.analyser = analyser
        self._blob_root = os.path.join(self.analyser.blob_root, 'state')

    def export(self, frame: int):
        output_dir = os.path.join(self._blob_root, f'frame_{frame}')
        os.makedirs(output_dir, exist_ok=True)
        self._export_non_filament(output_dir, frame)
        self._export_filaments(output_dir, frame)
        self._export_links(output_dir, frame)

    def _export_non_filament(
            self,
            output_dir: str,
            frame: int
    ):
        coords = self.analyser.get_trajectories_non_filament(minimum_image=True)[frame]
        np.save(os.path.join(output_dir, 'coordinates_non_filament_particles.npy'), coords)

    def _export_filaments(self, output_dir: str, frame: int):
        coords = self.analyser.get_positions_of_filaments_at_frames([frame], minimum_image=True)[0]
        np.save(os.path.join(output_dir, 'coordinates_filaments.npy'), coords)

    def _export_links(self, output_dir: str, frame: int):
        links = self.analyser.get_links(frame)
        np.save(os.path.join(output_dir, 'links.npy'), links)



import logging
import os
from multiprocessing import Pool
from typing import List, Tuple, Union, Optional

import h5py
import numpy as np
import matplotlib.pyplot as plt
from actomyosin_analyser.tools.group_results_container import GroupResultsContainer
from scipy.stats import gaussian_kde

from ._pipeline import Pipeline
from ..tools.experiment_configuration import ExperimentConfiguration
from ..tools.experiment_iterator import GroupIterator, Simulation
from ..analysis import geometry

logger = logging.getLogger(__name__)
_KDE_XY_Pair = Tuple[np.ndarray, np.ndarray]


class LengthDistribution(Pipeline):

    def _validate_configuration(self):
        assert "SegmentLength" in self.experiment_configuration

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 n_processes: int = 1):
        super().__init__(experiment_configuration)
        self._n_processes = n_processes

        self.plot_files.update({
            "length_distributions": os.path.join(experiment_configuration["SegmentLength"],
                                                 'length_distributions_{frames}.svg')
        })
        self.output_files.update({
            "length_distributions": os.path.join(experiment_configuration["SegmentLength"],
                                                 'length_distributions.h5')
        })
        self._lengths_folder = os.path.join(experiment_configuration['SegmentLength'], 'blobs')
        os.makedirs(self._lengths_folder, exist_ok=True)

    def run_analysis(self, n_points_kde: int,
                     frames: List[int]):
        density_estimates_list = self.get_density_estimates(n_points_kde, frames)
        fig, ax = plt.subplots(len(frames), 1, sharex=True)
        self.plot_density_estimates(ax, density_estimates_list, frames)
        fig.savefig(self.plot_files["length_distributions"].format(
            frames='_'.join([str(f) for f in frames])
        ))

    def make_violin_plot(self, frames: List[int], key: str) -> Tuple[plt.Figure, plt.Axes]:
        lengths = self.get_lengths(frames)
        index = self.experiment_configuration.experiment_iterator.create_index_from_groups()

        data_arrays = []
        labels = []
        for idx in sorted(index.get_level_values(key).unique()):
            selected_groups = self.experiment_configuration.experiment_iterator.get(**{key: idx})
            if len(selected_groups) > 1:
                raise RuntimeError(f"More than one group were selected with query {key}={idx}.")
            group = selected_groups[0]
            values = lengths[group]
            data_arrays.append(values)
            labels.append(str(idx))

        fig, ax = plt.subplots(1, 1)
        ax.violinplot(data_arrays, showmeans=True)
        ax.set_xticks(list(range(1, len(labels)+1)))
        ax.set_xticklabels(labels)
        return fig, ax

    def get_lengths(self, frames: List[int]) -> GroupResultsContainer[np.ndarray]:
        container = GroupResultsContainer[np.ndarray]()
        iterator = self.experiment_configuration.experiment_iterator
        for group in iterator:
            lengths = []
            for f in frames:
                lengths_f = self.get_lengths_by_group(group, f)
                lengths.append(lengths_f)
            container[group] = np.concatenate(lengths)
        return container

    def plot_density_estimates(self, ax, density_estimates: List[List[_KDE_XY_Pair]],
                               frames: List[int]):
        for i, f in enumerate(frames):
            self._plot_distributions_single_frame(ax[i], density_estimates[i])
            if i == 0:
                ax[i].legend()
            ax[i].set(
                title=f'frame {f}',
                ylabel='density',
            )
            if i + 1 == len(frames):
                ax[i].set_xlabel('segment length / $x_0$')

    def get_density_estimates(
            self,
            n_points_kde: int,
            frames: List[int]
    ) -> List[List[_KDE_XY_Pair]]:

        result_xy_pairs = []
        for f in frames:
            result_xy_pairs.append(self._get_density_estimates_single_frame(n_points_kde, f))

        return result_xy_pairs

    def _get_density_estimates_single_frame(
            self,
            n_points_kde: int,
            frame: int
    ) -> List[_KDE_XY_Pair]:
        iterator = self.experiment_configuration.experiment_iterator
        result_xy_pairs = []
        for group in iterator:
            x, density_estimate = self._get_density_estimates_single_group(group, n_points_kde, frame)
            result_xy_pairs.append((x, density_estimate))

        return result_xy_pairs

    def _get_density_estimates_single_group(
            self,
            group: GroupIterator,
            n_points_kde: int,
            frame: int
    ) -> _KDE_XY_Pair:

        label = group.get_label_from_values()
        x, density_estimate = self._load_density_estimate(label, n_points_kde, frame)
        if x is not None:
            return x, density_estimate
        x, density_estimate = self._compute_density_estimate(group, n_points_kde, frame)
        self._save_density_estimate(x, density_estimate,
                                    label, n_points_kde, frame)
        return x, density_estimate

    def _load_density_estimate(
            self,
            label: str,
            n_points_kde: int,
            frame: int
    ) -> Union[_KDE_XY_Pair, Tuple[None, None]]:
        fname = self.output_files["length_distributions"]
        if not os.path.exists(fname):
            return None, None

        with h5py.File(fname, 'r') as file:
            if label not in file:
                return None, None
            g = file[label]
            if str(frame) not in g:
                return None, None
            gg = g[str(frame)]
            if str(n_points_kde) not in gg:
                return None, None
            ggg = gg[str(n_points_kde)]

            x = ggg['x'][:]
            de = ggg['density_estimate'][:]
            return x, de

    def _compute_density_estimate(
            self,
            group: GroupIterator,
            n_points_kde: int,
            frame: int
    ) -> _KDE_XY_Pair:
        lengths = self.get_lengths_by_group(group, frame)
        logger.warning("Saving lengths to lengths.npy, this is a debugging add-on."
                       " Remove it when bug is resolved!!")
        np.save('lengths.npy', lengths)
        kde = gaussian_kde(lengths)
        x = np.linspace(lengths.min(), lengths.max(), n_points_kde)
        return x, kde(x)

    def get_lengths_by_group(
            self,
            group: GroupIterator,
            frame: int
    ) -> np.ndarray:
        lengths = self._load_lengths(group, frame)
        if lengths is not None:
            return lengths
        lengths = self._compute_lengths(group, frame)
        self._save_lengths(lengths, group, frame)
        return lengths

    def _load_lengths(
            self,
            group: GroupIterator,
            frame: int
    ) -> Optional[np.ndarray]:
        file_name = self._get_lengths_file_name(frame, group)
        if not os.path.isfile(file_name):
            return None
        return np.load(file_name)

    def _save_lengths(
            self,
            lengths: np.ndarray,
            group: GroupIterator, frame: int
    ):
        file_name = self._get_lengths_file_name(frame, group)
        np.save(file_name, lengths)

    def _get_lengths_file_name(self, frame, group) -> str:
        file_name = os.path.join(
            self._lengths_folder,
            f'lengths_{group.get_label_from_values()}_{frame}.npy'
        )
        return file_name

    def _compute_lengths(
            self,
            group: GroupIterator,
            frame: int
    ):
        selected_simulations = [sim for sim in group]
        pool = Pool(self._n_processes)

        caller = _GetLengthsCaller(frame)

        lengths = pool.map(
            caller,
            selected_simulations
        )

        lengths = np.concatenate(lengths).flatten()
        return lengths

    def _save_density_estimate(
            self,
            x: np.ndarray,
            density_estimate: np.ndarray,
            label: str,
            n_points_kde: int,
            frame: int
    ):
        fname = self.output_files["length_distributions"]
        with h5py.File(fname, 'a') as file:
            if label in file:
                g = file[label]
            else:
                g = file.create_group(label)
            if str(frame) in g:
                gg = g[str(frame)]
            else:
                gg = g.create_group(str(frame))
            if str(n_points_kde) in gg:
                ggg = gg[str(n_points_kde)]
            else:
                ggg = gg.create_group(str(n_points_kde))

            ggg['x'] = x
            ggg['density_estimate'] = density_estimate

    def _plot_distributions_single_frame(
            self,
            ax: plt.Axes,
            density_estimate_pairs: List[_KDE_XY_Pair]
    ):
        iterator = self.experiment_configuration.experiment_iterator
        for i, group in enumerate(iterator):
            x, estimate = density_estimate_pairs[i]
            ax.plot(x, estimate, label=group.latex_label, color=group.color)


class _GetLengthsCaller:

    def __init__(self, frame: int):
        self.frame = frame

    def __call__(self, sim: Simulation) -> np.ndarray:
        return self._get_lengths_single_simulation(sim)

    def _get_lengths_single_simulation(
            self,
            simulation: Simulation
    ) -> np.ndarray:
        lengths = []
        coords = simulation.analyser.get_trajectories_filaments(minimum_image=False)[self.frame]
        filaments = simulation.analyser.get_filaments()[self.frame]

        box = simulation.analyser.data_reader.read_box_size()
        box = np.array([box[1, 0] - box[0, 0],
                        box[1, 1] - box[0, 1],
                        box[1, 2] - box[0, 2]])

        for f in filaments:
            fcoords = coords[f.items]
            v_segments = geometry.get_minimum_image_vector(
                fcoords[:-1], fcoords[1:], box)
            d = np.sqrt(np.sum(v_segments ** 2, 1))
            lengths.append(d)
        return np.array(lengths)



import numpy as np
import glob


class SegmentDataReader:

    def __init__(self, folder):
        self.files = glob.glob(folder + "actin_coords_*.txt")
        self.files.sort()
        if len(self.files) == 0:
            raise RuntimeError("No actin coordinates in folder '" + folder + "'")
        self.lines_in_files = np.ones(len(self.files), dtype=np.int64)
        self.lines_in_files = self.lines_in_files * -1
        self._open_file(0)

    def _open_file(self, i):
        if i >= len(self.files):
            raise RuntimeError("index " + str(i) + " exceeds number of files")
        for j in range(i):
            if (self.lines_in_files[j] == -1):
                message = "Trying to open file " + str(i) + ", but file "\
                          + str(j) + " has never been opened.\n"
                message += "Files need to be opened sequentially to get"
                message += "the mapping of the time_step right."
                raise RuntimeError(message)
        self.array_current_file = np.genfromtxt(self.files[i])
        if self.array_current_file.ndim == 1:
            self.array_current_file = self.array_current_file.reshape(
                self.array_current_file.shape + (1,))
        self.current_file = i
        self.lines_in_files[i] = self.array_current_file.shape[0]
        self.first_time_step_of_current_array = np.sum(self.lines_in_files[:i])

    def get_coordinates_at_time_step(self, time_step):
        file_index = self._locate_file_containing_time_step(time_step)
        if (file_index == -1):
            raise RuntimeError("time_step " + str(time_step) + " is out of range")
        if self.current_file != file_index:
            self._open_file(file_index)
        return self.array_current_file[time_step - self.first_time_step_of_current_array, :]

    def _locate_file_containing_time_step(self, time_step):
        _sum_lines = 0
        for i in range(len(self.lines_in_files)):
            if (self.lines_in_files[i] == -1):
                self._open_file(i)
            _sum_lines = np.sum(self.lines_in_files[:i + 1])
            if (_sum_lines > time_step):
                return i
        return -1



import numba
import numpy as np
from numba.core import types
from numba.typed import Dict as NumbaDict

from actomyosin_analyser.analysis import geometry
from actomyosin_analyser.analysis.abstract_analyser import AbstractAnalyser


def _compute_filament_center_of_mass(
        filament_coordinates: np.ndarray,
        box: np.ndarray
) -> np.ndarray:
    relative_coords = geometry.get_minimum_image_vector(
        filament_coordinates,
        filament_coordinates[0],
        box
    )
    com = relative_coords.mean(0) + filament_coordinates[0]
    return com


class FilamentCenterOfMassGetter:

    def __init__(self, analyser: AbstractAnalyser):
        self.analyser = analyser

    def get(self) -> 'FilamentCenterOfMassTrajectories':
        coordinates = self.analyser.get_trajectories_filaments(minimum_image=False)
        filaments = self.analyser.get_filaments()
        simulation_box = self.analyser.read_simulation_box()

        com_trajectories = FilamentCenterOfMassTrajectories(coordinates.shape[0])
        for t, filaments_t in enumerate(filaments):
            coords_t = coordinates[t]
            for fil in filaments_t:
                coords_fil = coords_t[fil.items]
                center_of_mass = _compute_filament_center_of_mass(
                    coords_fil,
                    simulation_box
                )
                com_trajectories.set(fil.id, t, center_of_mass)
        return com_trajectories


class FilamentCenterOfMassTrajectories:

    def __init__(self, n_frames: int):
        self._n_frames = n_frames
        self._trajectories = NumbaDict.empty(types.int64, types.float64[:, :])

    def set(self, filament_id: int, frame: int, position: np.ndarray):
        _set_entry_in_trajectory_dict(self._trajectories, self._n_frames, filament_id, frame, position)

    def as_array(self) -> np.ndarray:
        return _trajectory_dict_to_array(self._trajectories, self._n_frames)


@numba.njit
def _trajectory_dict_to_array(d, n_frames) -> np.ndarray:
    trajectory_array = np.full((n_frames, len(d), 3), np.nan)
    for i, fil_id in enumerate(sorted(d.keys())):
        trajectory_array[:, i] = d[fil_id]
    return trajectory_array


@numba.njit()
def _set_entry_in_trajectory_dict(trajectory_dict, n_frames, filament_id, frame, position):
    if filament_id not in trajectory_dict:
        new_trajectory = np.full((n_frames, 3), np.nan)
        trajectory_dict[filament_id] = new_trajectory
    trajectory_dict[filament_id][frame] = position


from typing import List, Iterable, Union, Tuple, Dict, Any
import os
from math import isclose
import numpy as np
from scipy.spatial.transform import Rotation
from ..analysis.analyser import Analyser

FilamentID = int
BeadID = int
XYZTuple = Tuple[int, float, float, float]

class XYZExporter:

    def __init__(self, particle_type: str=None):
        self.particle_type = particle_type
        self.selected_filaments = None
        self.selected_beads = None
        self.analyser = None
        self.fname_template = None
        self.output_folder = None
        self.highlighted_beads = None
        self.CoM = False

    @staticmethod
    def export_all(
            analyser: Analyser,
            folder: str,
            fiber_radius: float=0.1,
            name_non_filament_particles='non_filament',
            name_links='link',
            frames: Union[List[int], None]=None
    ):
        trajectories_filaments = analyser.get_trajectories_filaments(minimum_image=True)
        if frames is None:
            frames = np.arange(len(trajectories_filaments))

        os.makedirs(folder, exist_ok=True)

        fname_fibers = os.path.join(folder, 'fibers.xyz')
        fname_non_fil = os.path.join(folder, name_non_filament_particles + '.xyz')
        fname_links = os.path.join(folder, name_links + '.xyz')

        XYZExporter.export_fibers(analyser, fname_fibers, fiber_radius,
                                  trajectories_filaments, frames)
        XYZExporter.export_non_filament_particles(analyser, fname_non_fil,
                                                  name_non_filament_particles, frames)
        XYZExporter.export_links(analyser, fname_links, name_links,
                                 trajectories_filaments,
                                 radius=fiber_radius*1.5,
                                 frames=frames)

    @staticmethod
    def export_fibers(
            analyser: Analyser,
            fname: str,
            fiber_radius: float,
            trajectories: np.ndarray,
            frames: List[int]
    ):
        
        fh = open(fname, 'wt')

        trajectories = trajectories
        filaments_all = analyser.data_reader.get_filaments_all()
        periodic_box = analyser.data_reader.read_box_size()
        periodic_box = periodic_box[1] - periodic_box[0]

        for f in frames:
            filaments = filaments_all[f]
            traj_t = trajectories[f]

            n_particles = 0
            block_xyz = ""
        
            for fil in filaments:
                n, b = XYZExporter._filament_to_xyz(traj_t[fil.items], periodic_box, fiber_radius)
                block_xyz += b
                n_particles += n

            block_xyz = "{}\n\n".format(n_particles) + block_xyz
            fh.write(block_xyz)
            
        fh.close()

    @staticmethod
    def export_non_filament_particles(
            analyser: Analyser,
            fname: str,
            label: str,
            frames: List[int],
            particle_range: Tuple[int, int]=None,
    ):
        trajectories = analyser.get_trajectories_non_filament(minimum_image=True)
        if particle_range is not None:
            trajectories = trajectories[:, particle_range[0]:particle_range[1]]

        fh = open(fname, 'wt')

        for f in frames:
            traj_f = trajectories[f]
            n_particles = len(traj_f)
            block_xyz = ""
            for p in traj_f:
                block_xyz += f"{label} {p[0]:.2f} {p[1]:.2f} {p[2]:.2f}\n"

            block_xyz = "{}\n\n".format(n_particles) + block_xyz
            fh.write(block_xyz)
        
        fh.close()

    @staticmethod
    def export_links(analyser: Analyser, fname: str, label: str, 
                     trajectories: np.ndarray, radius: float, frames: List[int]):

        periodic_box = analyser.data_reader.read_box_size()
        periodic_box = periodic_box[1] - periodic_box[0]

        links = analyser.data_reader.get_links_all()

        fh = open(fname, 'wt')

        for f in frames:
            traj_t = trajectories[f]
            links_t = links[f]

            w = np.where(links_t[:, 2] != -1)[0]
            pairs = []
            skip = []
            for wi in w:
                if wi in skip:
                    continue
                other = links_t[wi, 2]
                assert other in w
                pairs.append((wi, other))
                skip.append(other)

            n_particles = 0
            block_xyz = ""
            for p in pairs:
                n, b = XYZExporter._filament_to_xyz(traj_t[(p,)], periodic_box, radius,
                                                    label)
                block_xyz += b
                n_particles += n

            if len(pairs) == 0:
                n_particles = 1
                line = "{} {:.2f} {:.2f} {:.2f} {:.4f} {:.4f} {:.4f} {:.4f} {:.2f} {:.2f}\n"
                block_xyz += line.format(
                    'dummy', *([0.0]*9))

            block_xyz = "{}\n\n".format(n_particles) + block_xyz
            fh.write(block_xyz)

        fh.close()
                     
    @staticmethod
    def _filament_to_xyz(coords: np.ndarray, periodic_box: np.ndarray,
                         fiber_radius: float,
                         label: str=None) -> Tuple[int, str]:

        if label is None:
            label_tail = 'tail'
            label_core = 'core'
            label_head = 'head'
        else:
            label_tail = label
            label_core = label
            label_head = label

        count = 0
        xyz = ""

        shifted_coords = coords + periodic_box/2
        
        shifted_coord_sets = split_filament_coords_at_borders(shifted_coords, periodic_box)

        for i in range(len(shifted_coord_sets)):
            c = shifted_coord_sets[i] - periodic_box/2                

            centers = (c[1:] + c[:-1]) * 0.5
            segments = c[1:] - c[:-1]
            slengths = np.sqrt((segments**2).sum(1))
            orientations = segments / slengths[:, np.newaxis]

            xyz_c = ""
        
            line = "{} {:.2f} {:.2f} {:.2f} {:.4f} {:.4f} {:.4f} {:.4f} {:.2f} {:.2f}\n"
            specifier = np.array([label_core]*len(centers))
            if i == 0:
                specifier[0] = label_tail
            if i == len(shifted_coord_sets)-1:
                specifier[-1] = label_head

            for j in range(len(centers)):
                theta, phi = get_angles(orientations[j])
                r = Rotation.from_euler('yz', [theta, phi])
                quat = r.as_quat()

                xyz_c += line.format(
                    specifier[j],
                    centers[j, 0],
                    centers[j, 1],
                    centers[j, 2],
                    quat[0],
                    quat[1],
                    quat[2],
                    quat[3],
                    fiber_radius,
                    slengths[j]
                )
            count += centers.shape[0]
            xyz += xyz_c
        return count, xyz
        
    def export(self, analyser,
               t: Union[int, Iterable[int]],
               selected_filaments: Iterable[FilamentID]=None,
               selected_beads: Iterable[BeadID]=None,
               highlighted_beads: Union[Dict[int, List[BeadID]],
                                        List[BeadID]]=None,
               label: str=None,
               CoM: bool=False):
        raise NotImplementedError('needs to be updated for new analyser (not rwfilaments_analyser)')
        self.analyser = analyser        
        self.output_folder = os.path.join(analyser.dirname, 'xyz_export')
        if not os.path.exists(self.output_folder):
            os.mkdir(self.output_folder)
        if label is not None:
            self.output_folder = os.path.join(self.output_folder, label)
            if not os.path.exists(self.output_folder):
                os.mkdir(self.output_folder)
        self.selected_beads = selected_beads
        self.selected_filaments = selected_filaments
        self.highlighted_beads = highlighted_beads
        self.CoM = CoM

        if isinstance(t, int):
            self.fname_template = 'beads.xyz'
            self._export_single_file(t)
        else:
            t_max = t[-1]
            self.fname_template = f'beads_{{t:0{len(str(t_max))}}}.xyz'
            out_str = f"exporting for t = {{ti:0{len(str(t_max))}}}/{{t_max:0{len(str(t_max))}}}"
            for ti in t:
                print(out_str.format(ti=ti, t_max=t_max), end='\r')
                self._export_single_file(ti)
            print()

        self.selected_filaments = None
        self.output_folder = None
        self.selected_beads = None
        self.analyser = None
        self.fname_template = None
        self.highlighted_beads = None
        self.CoM = False

    def _get_xyz_data(self, t: int) -> List[XYZTuple]:
        if self.particle_type is not None:
            raise RuntimeWarning("particle_type parameter does not do anything yet!")
        xyz_data = []
        coords = self.analyser.get_coordinates(True)[t]
        if self.CoM:
            coords = coords - coords.mean(0)
        indices = self._get_selected_beads(t)
        bids = self.analyser.data_reader.data_set_bead_ids[t]
        for idx in indices:
            ci = coords[idx]
            ptype = self._get_particle_type(bids[idx])
            xyz_data.append((ptype, ci[0], ci[1], ci[2]))
        
        return xyz_data        
        
    def _get_particle_type(self, bid: BeadID) -> int:
        if self.highlighted_beads is None:
            return 0
        if isinstance(self.highlighted_beads, list):
            if bid in self.highlighted_beads:
                return 1
            return 0

        for k in self.highlighted_beads:
            if bid in self.highlighted_beads[k]:
                return k
        return 0

    def _get_selected_beads(self, t: int) -> np.ndarray:
        bids = self.analyser.data_reader.data_set_bead_ids[t]
        if self.selected_filaments is None and self.selected_beads is None:
            return np.arange(len(bids))[bids != -1]
        indices = set()
        if self.selected_beads is not None:
            for idx, bid in enumerate(bids):
                if bid in self.selected_beads:
                    indices.add(idx)
        if self.selected_filaments is None:
            return np.array(indices)

        fidxerd = self.analyser.get_filament_indexer_dict()

        for fid in self.selected_filaments:
            fidxer = fidxerd[fid][t]
            indices.update(fidxer)

        return np.array(list(indices))

    def _export_single_file(self, t: int):
        f = open(os.path.join(self.output_folder, self.fname_template.format(t=t)), 'wt')
        xyz_data = self._get_xyz_data(t)
        f.write(f'{len(xyz_data)}\n\n')
        for xyz_i in xyz_data:            
            f.write('{} {} {} {}\n'.format(*xyz_i))
        f.close()


def get_angles(v: np.ndarray) -> Tuple[float, float]:
    """
    Get angles theta, phi in spherical coordinates of vector v.
    """
    theta = np.arccos(v[2])
    # phi = np.arccos(v[0]/np.sin(theta))
    phi = np.arctan2(v[1], v[0])
    return theta, phi


def minimum_image_projection(
        coords: np.ndarray,
        box: Union[np.ndarray, List[float]]) -> np.ndarray:
    proj = np.empty_like(coords)
    for i in range(len(box)):
        proj[:, i] = coords[:, i] -\
            np.floor_divide(coords[:, i], box[i]) * box[i]
    return proj


def split_filament_coords_at_borders(ordered_bead_coords: np.ndarray,
                                     box) -> List[np.ndarray]:
    """
    @param ordered_bead_coords Coordinates of beads, have to be corrected via minimum image!
    """
    coordinate_sets = []
    diff = ordered_bead_coords[1:] - ordered_bead_coords[:-1]
    shifts = _compute_shifts_array(diff, box)
    shifts_at = np.where(shifts != 0)
    if len(shifts_at[0]) == 0:
        return [ordered_bead_coords]
    shifts_at_n, shifts_at_i, r_cuts = _sort_conflicting_shifts(shifts_at[0], shifts_at[1],
                                                                ordered_bead_coords, diff,
                                                                shifts, box)
    shifted_diffs = diff + shifts
    n_prev = shifts_at_n[0]
    prev_virtual_bead = None

    coord_set, prev_virtual_bead = _generate_coord_set_before_first_split(
        ordered_bead_coords[:shifts_at_n[0] + 1], n_prev, r_cuts[0], shifted_diffs)
    coordinate_sets.append(coord_set)
    for j, n in enumerate(shifts_at_n[1:]):
        j = j+1
        set_shape = list(ordered_bead_coords.shape)
        if n == n_prev:
            num_points = 2
        else:
            num_points = n - n_prev + 2
        set_shape[0] = num_points
        coord_set = np.empty(tuple(set_shape))
        if prev_virtual_bead is not None:
            prev_i = shifts_at_i[j-1]
            if isclose(prev_virtual_bead[prev_i], box[prev_i], abs_tol=box[prev_i] * 0.01):
                prev_virtual_bead[prev_i] -= box[prev_i]
            else:
                prev_virtual_bead[prev_i] += box[prev_i]
            coord_set[0] = prev_virtual_bead
        if n == n_prev:
            virtual_bead = coord_set[0] + (r_cuts[j] - r_cuts[j-1]) * shifted_diffs[n]
        else:
            coord_set[1: -1] = ordered_bead_coords[n_prev+1: n+1]
            virtual_bead = ordered_bead_coords[n] + r_cuts[j] * shifted_diffs[n]
        prev_virtual_bead = virtual_bead
        coord_set[-1] = virtual_bead

        coordinate_sets.append(coord_set)
        n_prev = n
        
    coord_set = _generate_coord_set_after_last_split(
        ordered_bead_coords, n_prev, shifts_at_i, prev_virtual_bead, box)
    
    coordinate_sets.append(coord_set)
    
    return coordinate_sets


def _generate_coord_set_before_first_split(bead_coords_before_split, n, r_cut, shifted_diffs):
    coord_set_shape = list(bead_coords_before_split.shape)
    coord_set_shape[0] = coord_set_shape[0] + 1
    coord_set = np.empty(tuple(coord_set_shape))
    coord_set[:-1] = bead_coords_before_split[:]
    virtual_bead = bead_coords_before_split[-1] + r_cut * shifted_diffs[n]
    coord_set[-1] = virtual_bead
    return coord_set, virtual_bead
    

def _generate_coord_set_after_last_split(ordered_bead_coords, n_prev,
                                         shifts_at_i, prev_virtual_bead, box):
    num_points = len(ordered_bead_coords) - n_prev
    set_shape = list(ordered_bead_coords.shape)
    set_shape[0] = num_points
    coord_set = np.empty(tuple(set_shape))
    last_i = shifts_at_i[-1]
    if isclose(prev_virtual_bead[last_i], box[last_i], abs_tol=box[last_i] * 0.01):
        prev_virtual_bead[last_i] -= box[last_i]
    else:
        prev_virtual_bead[last_i] += box[last_i]
    coord_set[0] = prev_virtual_bead
    coord_set[1:] = ordered_bead_coords[n_prev+1:]
    return coord_set

def _compute_shifts_array(diff, box):
    shifts = np.zeros_like(diff)
    for i, b in enumerate(box):
        if b is None:            
            continue
        shifts_i = -b * np.rint(diff[:, i] / b)
        shifts[:, i] = shifts_i
    return shifts
        
def _sort_conflicting_shifts(shifts_at_n, shifts_at_i,
                             ordered_bead_coords, diff, shifts, box):

    r_cuts = np.zeros(shifts_at_n.shape)
    for j in range(len(shifts_at_n)):
        conflicting_indices = []
        nj = shifts_at_n[j]
        k = j + 1
        while k < len(shifts_at_n):
            nk = shifts_at_n[k]
            if nj != nk:
                break
            conflicting_indices.append(k)
            k += 1
        if len(conflicting_indices) == 0:
            ij = shifts_at_i[j]
            dx = diff[nj, ij] + shifts[nj, ij]
            x_nj_ij = ordered_bead_coords[nj, ij]
            if x_nj_ij + dx > box[ij]:
                r_cut = abs((box[ij] - x_nj_ij) / dx)
            else:                
                r_cut = abs(x_nj_ij / dx)
            r_cuts[j] = r_cut
            continue
        conflicting_indices.insert(0, j)
        boundary_cuts = {}
        for l in conflicting_indices:
            nl = shifts_at_n[l]
            il = shifts_at_i[l]
            dx = diff[nl, il] + shifts[nl, il]
            x_nl_il = ordered_bead_coords[nl, il]
            if x_nl_il + dx > box[il]:
                r_cut = abs((box[il] - x_nl_il) / dx)
            else:                
                r_cut = abs(x_nl_il / dx)
            boundary_cuts[l] = r_cut
        sorted_boundary_cut_pairs = sorted(boundary_cuts.items(), key=lambda x: x[1])

        conflicting_shifts_in_correct_order_n = []
        conflicting_shifts_in_correct_order_i = []
        r_cuts_in_correct_order = []
        # a cut pair is a Tuple of the index in the shifts_at array and the r_cut value
        for cut_pair in sorted_boundary_cut_pairs:
            conflicting_shifts_in_correct_order_n.append(shifts_at_n[cut_pair[0]])
            conflicting_shifts_in_correct_order_i.append(shifts_at_i[cut_pair[0]])
            r_cuts_in_correct_order.append(cut_pair[1])
        shifts_at_n[j:j+len(conflicting_indices)] = conflicting_shifts_in_correct_order_n
        shifts_at_i[j:j+len(conflicting_indices)] = conflicting_shifts_in_correct_order_i
        r_cuts[j:j+len(conflicting_indices)] = r_cuts_in_correct_order
    return shifts_at_n, shifts_at_i, r_cuts



import os
from dataclasses import dataclass
from typing import Sequence, Union, Tuple

import numpy as np
import pandas as pd

from actomyosin_analyser.analysis.abstract_analyser import AbstractAnalyser


class PolymerDensityGetter:

    def __init__(
            self,
            grid: 'GridGenerator',
            analyser: AbstractAnalyser
    ):
        self._analyser = analyser
        self._results_dir = os.path.join(analyser.blob_root, 'polymer_density')
        os.makedirs(self._results_dir, exist_ok=True)
        self.grid_generator = grid

    def get(
            self,
            frames: Sequence[int],
            bandwidth: Union[float, None]
    ) -> Tuple[float, np.ndarray]:
        denstities = []
        for f in frames:
            dens, bandwidth = self._get_at_frame(f, bandwidth)
            denstities.append(dens[np.newaxis])
        return bandwidth, np.concatenate(denstities)

    def _get_at_frame(
            self,
            frame: int,
            bandwidth: Union[float, None]
    ) -> Tuple[np.ndarray, float]:
        if bandwidth is None:
            bandwidth = self._load_available_bandwidth(frame)
        density = self._load_density(frame, bandwidth)
        if density is not None:
            return density, bandwidth
        density, bandwidth = self._compute_density(frame, bandwidth)
        bandwidth = PolymerDensityGetter._emulate_loss_by_writing_and_rereading(bandwidth)
        self._save_bandwidth(frame, bandwidth)
        self._save_density(frame, density, bandwidth)
        return density, bandwidth

    def _load_density(self, frame: int, bandwidth: Union[None, float]) -> Union[None, np.ndarray]:
        if bandwidth is None:
            return None
        data_file_name = self._construct_data_file_name(frame, bandwidth)
        if not os.path.exists(data_file_name):
            return None
        return np.load(data_file_name)

    def _compute_density(
            self,
            frame: int,
            bandwidth: Union[None, float]
    ) -> Tuple[np.ndarray, float]:
        kde = self._analyser.get_densities_gaussian_kde_scipy(frame, bandwidth)
        bandwidth = kde.factor
        X, Y, Z = self.grid_generator.generate()
        positions = np.vstack([X.ravel(), Y.ravel(), Z.ravel()])
        dens = kde(positions)
        dens = dens.reshape(X.shape)
        return dens, bandwidth

    def _save_bandwidth(
            self,
            frame: int,
            bandwidth: float
    ) -> None:
        folder_name, file_name = self._construct_bandwidth_paths(
            frame, hash(bandwidth)
        )
        os.makedirs(folder_name, exist_ok=True)
        np.save(file_name, np.array([bandwidth]))
        self._add_entry_to_bandwidth_table(frame, bandwidth)

    def _construct_bandwidth_paths(
            self,
            frame: int,
            bandwidth_hash: int
    ) -> Tuple[str, str]:
        folder = self._construct_data_folder(bandwidth_hash, frame)
        file = os.path.join(folder, 'bandwidth.npy')
        return folder, file

    def _construct_data_folder(self, bandwidth_hash: int, frame):
        grid_folder = self._construct_grid_folder()
        folder = os.path.join(
            grid_folder,
            f'bandwidth_hash_{bandwidth_hash}',
            f'frame_{frame}'
        )
        return folder

    def _construct_grid_folder(self) -> str:
        return os.path.join(
            self._results_dir,
            f'grid_hash_{hash(self.grid_generator)}'
        )

    def _add_entry_to_bandwidth_table(
            self,
            frame: int,
            bandwidth: float
    ) -> None:
        grid_folder = self._construct_grid_folder()
        file = os.path.join(grid_folder, 'bandwidth_table.csv')
        if os.path.exists(file):
            table = pd.read_csv(file, index_col=0)
        else:
            table = pd.DataFrame(columns=['frame', 'bandwidth_hash'],
                                 dtype=int)
        index = 0 if len(table.index) == 0 else table.index.max()+1
        table.loc[index] = frame, hash(bandwidth)
        table.to_csv(file)

    def _save_density(
            self,
            frame: int,
            density: np.ndarray,
            bandwidth: float
    ) -> None:
        file_name = self._construct_data_file_name(frame, bandwidth)
        np.save(file_name, density)

    def _construct_data_file_name(
            self,
            frame: int,
            bandwidth: float
    ) -> str:
        data_folder = self._construct_data_folder(hash(bandwidth), frame)
        return os.path.join(
            data_folder, 'density.npy'
        )

    def _load_available_bandwidth(self, frame: int) -> Union[None, float]:
        bw_hash = self._get_bandwidth_hash_from_table(frame)
        if bw_hash is None:
            return None
        folder, file = self._construct_bandwidth_paths(frame, bw_hash)
        return np.load(file)[0]

    def _get_bandwidth_hash_from_table(self, frame: int) -> Union[None, int]:
        grid_folder = self._construct_grid_folder()
        file = os.path.join(grid_folder, 'bandwidth_table.csv')
        if not os.path.exists(file):
            return None
        table = pd.read_csv(file, index_col=0)
        at_frame = table[table['frame'] == frame]
        if len(at_frame) == 0:
            return None
        bw_hash = at_frame.iloc[0]['bandwidth_hash']
        return bw_hash

    @staticmethod
    def _emulate_loss_by_writing_and_rereading(bandwidth: float) -> float:
        tmp_file = '.tmp_bw.npy'
        np.save(tmp_file, np.array([bandwidth]))
        v = np.load(tmp_file)[0]
        os.remove(tmp_file)
        return v


@dataclass
class GridGenerator:
    """
    Handles setting up coordinates of a 3D grid.
    Spacings for each dimension have to be specified via tuples of three values:
    (minimum grid position, maximum grid position, number of equidistant
    points from min to max position (including min an max as outermost points)).
    """

    x_spacing: Tuple[float, float, int]
    y_spacing: Tuple[float, float, int]
    z_spacing: Tuple[float, float, int]

    def generate(self) -> Tuple[np.ndarray,
                                np.ndarray,
                                np.ndarray]:
        """
        Use ``numpy.mgrid`` to get arrays of X, Y, Z coordinates of points on the grid.

        :return: Tuple of 3 arrays: X, Y, Z coordinates of grid points.
        """
        X, Y, Z = np.mgrid[
                  self.x_spacing[0]: self.x_spacing[1]: self.x_spacing[2] * 1j,
                  self.y_spacing[0]: self.y_spacing[1]: self.y_spacing[2] * 1j,
                  self.z_spacing[0]: self.z_spacing[1]: self.z_spacing[2] * 1j
                  ]
        return X, Y, Z

    def __hash__(self):
        return hash((self.x_spacing, self.y_spacing, self.z_spacing))





import os
from typing import Tuple, List, Union
import logging

import h5py
import numpy as np

from actomyosin_analyser.tools.experiment_configuration import ExperimentConfiguration
from actomyosin_analyser.pipelines._pipeline import Pipeline
from actomyosin_analyser.tools.experiment_iterator import GroupIterator
from actomyosin_analyser.pipelines.length_distribution import LengthDistribution

logger = logging.getLogger(__name__)


class SegmentLength(Pipeline):

    def __init__(
            self,
            experiment_configuration: ExperimentConfiguration,
            n_processes: int=1
    ):
        super().__init__(experiment_configuration)
        self._ld_pipeline = LengthDistribution(experiment_configuration, n_processes)

        self.output_files.update({
            "segment_length": os.path.join(experiment_configuration['SegmentLength'],
                                           'segment_length.h5')
        })

    def run_analysis(self, frames: np.ndarray) -> List[Tuple[np.ndarray, np.ndarray]]:
        return self.get_means_and_variances(frames)


    def _validate_configuration(self):
        assert "SegmentLength" in self.experiment_configuration

    def get_means_and_variances(self, frames) -> List[Tuple[np.ndarray, np.ndarray]]:
        means_vars = []
        for group in self.experiment_configuration.experiment_iterator.groups:
            means_vars.append(self._get_means_and_variances_for_group(group, frames))

        return means_vars

    def _get_means_and_variances_for_group(
            self,
            group: GroupIterator,
            frames: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        means, variances = [], []
        for f in frames:
            logger.info(f"At group {group.get_label_from_values()}, "
                        f"frame {f}.")
            mean, var = self._get_mean_and_variance_for_frame(group, f)
            means.append(mean)
            variances.append(var)
        return np.array(means), np.array(variances)

    def _get_mean_and_variance_for_frame(self, group: GroupIterator, frame: int) -> Tuple[float, float]:
        mean, var = self._load_mean_and_variance(
            group, frame
        )
        if mean is not None:
            return mean, var
        mean, var = self._compute_mean_and_variance(group, frame)
        self._save_mean_and_variance(group, frame, mean, var)
        return mean, var

    def _compute_mean_and_variance(
            self,
            group: GroupIterator,
            frame: int
    ) -> Tuple[float, float]:
        lengths = self._ld_pipeline.get_lengths_by_group(group, frame)
        return lengths.mean(), lengths.var()

    def _save_mean_and_variance(self, group: GroupIterator, frame: int,
                                mean: float, variance: float):
        group_label = group.get_label_from_values()
        with h5py.File(self.output_files['segment_length'], 'a') as f:
            if group_label in f:
                h5_group = f[group_label]
            else:
                h5_group = f.create_group(group_label)
            frame_group = h5_group.create_group(str(frame))
            frame_group['mean'] = np.array([mean])
            frame_group['variance'] = np.array([variance])

    def _load_mean_and_variance(
            self,
            group: GroupIterator,
            frame: int
    ) -> Union[Tuple[None, None], Tuple[float, float]]:
        fname = self.output_files['segment_length']
        if not os.path.exists(fname):
            return None, None
        group_label = group.get_label_from_values()
        with h5py.File(fname, 'r') as f:
            if group_label not in f:
                return None, None
            h5_group = f[group_label]
            sf = str(frame)
            if sf not in h5_group:
                return None, None
            frame_group = h5_group[sf]
            return frame_group['mean'][0], frame_group['variance'][0]



from typing import List

STATE_FILAMENT_INTERIOR = 2.0
STATE_FILAMENT_START    = 1.0
STATE_FILAMENT_END      = 3.0


class Bead:

    def __init__(self):
        self.previous = -1
        self.next = -1
        self.cross = -1


class Filament:

    def __init__(self, filament_id: int, items: List[int], motors: List[int]):
        self.id = filament_id
        self.items = items
        self.motors = motors



from typing import List, Tuple
import os
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from ._g_pipeline import _GPipeline
from .bead_msd import BeadMSD, EnsembleMSDData
from ...tools.experiment_configuration import ExperimentConfiguration
from ...tools.experiment_iterator import GroupIterator
from ...analysis.microrheology.paust_method import (fit_msd,
                                                    paust_model,
                                                    compute_G_via_fit_parameters)
from ...file_io.tables import merge_and_save_table


class PaustMethod(_GPipeline):

    def __init__(self, experiment_configuration: ExperimentConfiguration,
                 overwrite: bool=False):
        super().__init__(experiment_configuration)
        self.plot_files.update({
            'PaustFit': os.path.join(experiment_configuration['G'], 'paust_fit_{label}.svg'),
            'G': os.path.join(experiment_configuration['G'], 'G.svg'),
            'G0': os.path.join(experiment_configuration['G0'], 'G_with_G0.svg')
        })
        self.output_files.update({
            'PaustFit': os.path.join(experiment_configuration['G'], 'paust_fit.csv'),
            'G0': os.path.join(experiment_configuration['G0'], 'G0.csv')
        })
        self._bead_msd = BeadMSD(experiment_configuration, overwrite)
        if overwrite:
            self._remove_output_files()

    def run_analysis(self) -> plt.Figure:
        msd_data = self._bead_msd.get_ensemble_msds()
        fig, ax = self._run_paust_analysis(msd_data)
        return fig, ax

    def _validate_configuration(self):
        assert "G" in self.experiment_configuration
        assert "G0" in self.experiment_configuration


    def get_msds_and_fit_results(self) -> Tuple['EnsembleMSDData', pd.DataFrame]:
        msd_data = self._bead_msd.get_ensemble_msds()
        fit_results = self._fit_paust_model_to_msds(msd_data)
        return msd_data, fit_results

    def _run_paust_analysis(
            self,
            msd_data: 'EnsembleMSDData'
    ) -> Tuple[plt.Figure, plt.Axes]:
        fit_results = self._fit_paust_model_to_msds(msd_data)
        self.plot_paust_fits(msd_data, fit_results)
        omegas, Gs = PaustMethod._get_omegas_and_Gs_from_fit_parameters(fit_results,
                                                                        msd_data.lag_times,
                                                                        msd_data.bead_radii)
        fig, ax = self._plot_Gs(omegas, Gs, save_file_name=self.plot_files['G'])
        G0s = self._find_G0_via_saddle_point_method(omegas, Gs, self.output_files["G0"])
        self._plot_Gs_with_G0s(omegas, Gs, G0s, self.plot_files['G0'])
        return fig, ax

    def get_omegas_and_Gs(self) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        msd_data = self._bead_msd.get_ensemble_msds()
        fit_results = self._fit_paust_model_to_msds(msd_data)
        omegas, Gs = PaustMethod._get_omegas_and_Gs_from_fit_parameters(fit_results,
                                                                        msd_data.lag_times,
                                                                        msd_data.bead_radii)
        return omegas, Gs

    def _fit_paust_model_to_msds(self, msd_data: 'EnsembleMSDData') -> pd.DataFrame:
        fit_results = self._create_fit_results_table()
        for i, group in enumerate(self.experiment_configuration.experiment_iterator):
            row = fit_results.iloc[i]
            self._check_table_values_match_group_values(group, row)
            lag_time, msd = msd_data.get_lag_time_and_msd(i)
            parameters, covariance_errors = fit_msd(lag_time, msd)
            p_errors = np.sqrt(np.diag(covariance_errors))
            fit_results.iloc[i] = np.concatenate([parameters, p_errors])
        merge_and_save_table(self.output_files['PaustFit'], fit_results)
        return fit_results.xs(self.skip, level='frames skipped')

    @staticmethod
    def _check_table_values_match_group_values(group: GroupIterator, row):
        group_values = list(tuple(group.values.values()))
        for i, gv in enumerate(group_values):
            if isinstance(gv, float):
                group_values[i] = np.round(gv, decimals=3)
        group_values = tuple(group_values)
        #if not (row.name[:-1] == group_value_array).all():
        #    raise RuntimeError("Order of fit results table does not match order"
        #                       " of groups.")
        if row.name[:-1] != group_values:
            raise RuntimeError("Order of fit results table does not match order"
                               " of groups.")

    def _create_fit_results_table(self) -> pd.DataFrame:
        index = self._create_group_index_with_skip_index()
        table = pd.DataFrame(index=index, columns=['A', 'B', 'C', 'D', 'sigma_A',
                                                   'sigma_B', 'sigma_C', 'sigma_D'])
        return table

    def plot_paust_fits(self, msd_data: 'EnsembleMSDData', fit_results: pd.DataFrame):
        groups = self.experiment_configuration.experiment_iterator.groups
        for i in range(msd_data.length):
            lag_time, msd = msd_data.get_lag_time_and_msd(i)
            fit_params = np.array(fit_results[['A', 'B', 'C', 'D']].iloc[i])
            self._plot_single_paust_fit(lag_time, msd, fit_params, groups[i])

    def _plot_single_paust_fit(self, lag_time: np.ndarray,
                               msd: np.ndarray,
                               fit_params,
                               group: GroupIterator):
        label = group.label
        file_label = group.get_label_from_values()
        fig, ax = plt.subplots(1, 1)
        filename = self.plot_files['PaustFit'].format(label=file_label)

        ax.plot(lag_time, msd, 'o')
        ax.plot(lag_time, paust_model(lag_time, *fit_params), label='Paust fit')
        ax.legend()
        ax.set(
            title=label,
            xlabel='$\\Delta t / t_0$',
            ylabel='MSD $\\langle \\Delta r^2 \\rangle / x_0^2$',
        )

        fig.savefig(filename)

    @staticmethod
    def _get_omegas_and_Gs_from_fit_parameters(
            fit_results: pd.DataFrame,
            lag_times: List[np.ndarray],
            bead_radii: List[float]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        omegas, Gs = [], []
        for i in range(len(bead_radii)):
            kT = 1
            bead_radius = bead_radii[i]
            omega = 1/lag_times[i]
            fit_parameters = np.array(fit_results[['A', 'B', 'C', 'D']].iloc[i])
            G = compute_G_via_fit_parameters(omega, bead_radius, kT, *fit_parameters)
            omegas.append(omega)
            Gs.append(G)
        return omegas, Gs



import os
from dataclasses import dataclass
from multiprocessing import Pool
from typing import Tuple, List, Sequence, Union

import numpy as np
from scipy.stats import gaussian_kde
import pandas as pd

from actomyosin_analyser.analysis.polymer_density import GridGenerator
from actomyosin_analyser.pipelines._pipeline import Pipeline
from actomyosin_analyser.tools.experiment_configuration import ExperimentConfiguration
from actomyosin_analyser.tools.experiment_iterator import GroupIterator, Simulation
from actomyosin_analyser.tools.group_results_container import GroupResultsContainer


class AveragePolymerDensity(Pipeline):

    def __init__(
            self,
            experiment_configuration: ExperimentConfiguration,
            number_of_processes: int
    ):
        super().__init__(experiment_configuration)

        self.output_files.update({
            "densities": os.path.join(experiment_configuration["ActinDensity"],
                                      "density.h5"),
            'bandwidth_table': os.path.join(experiment_configuration["ActinDensity"],
                                            "kde_bandwidths.h5")
        })
        self._n_processes = number_of_processes

    def run_analysis(
            self,
            frames: Sequence[int],
            x_grid_parameters: Tuple[float, float, int],
            y_grid_parameters: Tuple[float, float, int],
            z_grid_parameters: Tuple[float, float, int],
    ) -> '_DensityResults':
        """
        Computes actin density on specified grid. Gaussian kernel density
        estimate is used to estimate densities on the grid.
        Grid parameters have to be a tuple of starting point (float), ending point(float),
        and number of points (int) along the length from start to end.
        """
        return self.get_densities(
            frames,
            x_grid_parameters,
            y_grid_parameters,
            z_grid_parameters
        )

    def _validate_configuration(self):
        assert "ActinDensity" in self.experiment_configuration

    def get_densities(
            self,
            frames: Sequence[int],
            x_grid_parameters: Tuple[float, float, int],
            y_grid_parameters: Tuple[float, float, int],
            z_grid_parameters: Tuple[float, float, int],
            same_bandwidth_across_groups: bool=False
    ) -> '_DensityResults':
        """
        Computes actin density on specified grid. Gaussian kernel density
        estimate is used to estimate densities on the grid.
        Grid parameters have to be a tuple of starting point (float), ending point(float),
        and number of points (int) along the length from start to end.
        """
        gg = GridGenerator(x_grid_parameters, y_grid_parameters, z_grid_parameters)
        densities = _DensityResults()
        bandwidth = None
        bandwidth_hash = None

        groups = self.experiment_configuration.experiment_iterator.groups
        if same_bandwidth_across_groups:
            group = groups[0]
            dens = self.get_densities_of_group(group, gg, frames, bandwidth, bandwidth_hash)
            densities[group] = dens
            bandwidth = dens.bandwidth
            bandwidth_hash = dens.bandwidth_hash
            remaining_groups = groups[1:]
        else:
            remaining_groups = groups[:]

        pool = Pool(self._n_processes)

        average_density_getter = _AverageDensityGetter(
            self.experiment_configuration["ActinDensity"], frames,
            gg, bandwidth, bandwidth_hash
        )

        dens_results = pool.map(average_density_getter, remaining_groups)

        for i, g in enumerate(remaining_groups):
            densities[g] = dens_results[i]

        return densities

    def get_densities_of_group(
            self,
            group: GroupIterator,
            grid: 'GridGenerator',
            frames: Sequence[int],
            bandwidth: Union[None, float],
            bandwidth_hash: Union[None, int]
    ) -> 'KDEResult':
        adg = _AverageDensityGetter(
            self.experiment_configuration["ActinDensity"], frames,
            grid, bandwidth, bandwidth_hash
        )
        dens = adg(group)
        return dens


@dataclass
class KDEResult:
    densities: List[np.ndarray]
    bandwidth: float
    bandwidth_hash: int
    grid_generator: GridGenerator

    def __getitem__(self, index: int) -> np.ndarray:
        return self.densities[index]


@dataclass
class _AverageDensityGetter:
    results_folder: str
    frames: Sequence[int]
    grid_generator: GridGenerator
    bandwidth: Union[None, float]
    bandwidth_hash: Union[None, int]

    def __call__(self, group: GroupIterator) -> 'KDEResult':
        return self._get_average_densities_single_group(group)

    def _get_average_densities_single_group(
            self,
            group: GroupIterator,
    ) -> 'KDEResult':
        avg_densities = []

        bandwidth = self.bandwidth
        bandwidth_hash = self.bandwidth_hash
        for f in self.frames:
            avg_f, bandwidth, bandwidth_hash = self._get_average_densities_at_frame(
                group,
                f,
                bandwidth,
                bandwidth_hash
            )
            avg_densities.append(avg_f)
        return KDEResult(avg_densities, bandwidth, bandwidth_hash, self.grid_generator)

    def _get_average_densities_at_frame(
            self,
            group: GroupIterator,
            frame: int,
            bandwidth: float,
            bandwidth_hash: int
    ) -> Tuple[np.ndarray, float, int]:
        label = group.get_label_from_values()
        avg_densities, bandwidth, bandwidth_hash = self._load_average_densities(
            label, frame, bandwidth, bandwidth_hash
        )
        if avg_densities is not None:
            print(f"loaded average densities of group {label}")
            return avg_densities, bandwidth, bandwidth_hash
        avg_densities, bandwidth, bandwidth_hash = self._compute_average_densities(
            group, frame, bandwidth, bandwidth_hash
        )
        self._save_average_densities(avg_densities, label,
                                     frame,
                                     bandwidth, bandwidth_hash)
        return avg_densities, bandwidth, bandwidth_hash

    def _compute_average_densities(
            self,
            group: GroupIterator,
            frame: int,
            bandwidth: Union[float, None],
            bandwidth_hash: Union[int, None]
    ) -> Tuple[np.ndarray, float, int]:
        simulations = [sim for sim in group]
        kdes = []
        for i, sim in enumerate(simulations):
            print(f"Computing KDE for group {group.label}, frame {frame}"
                  f"bandwidth is {bandwidth}: simulation {i+1:02}/{len(simulations)}")
            kde = self._get_kde(sim, frame, bandwidth)
            if bandwidth is None:
                bandwidth = kde.factor
            if bandwidth_hash is None:
                bandwidth_hash = hash(bandwidth)
            kdes.append(kde)

        X, Y, Z = self.grid_generator.generate()

        average_densities = np.zeros_like(X)

        for j, sim in enumerate(simulations):
            print(f"computing density for group {group.label}, frame {frame}, simulation {j+1:02}/{len(simulations)}")
            dens = self._get_density_grid_single(kdes[j], X, Y, Z)
            average_densities += dens
        average_densities /= len(simulations)

        return average_densities, bandwidth, bandwidth_hash

    def _save_average_densities(
            self,
            avg_densities: np.ndarray,
            label: str,
            frame: int,
            bandwidth: float,
            bandwidth_hash: int
    ) -> None:
        self._add_entry_to_bandwidth_table(bandwidth, bandwidth_hash, label)
        file_name, folder_name = self._construct_folder_and_file_name(bandwidth_hash, frame, label)
        os.makedirs(folder_name, exist_ok=True)
        np.save(file_name, avg_densities)

    def _construct_folder_and_file_name(self, bandwidth_hash, frame, label):
        frame_label = f'frame_{frame}'
        grid_label = f"grid_hash_{hash(self.grid_generator)}"
        bandwidth_label = f"bandwidth_hash_{bandwidth_hash}"
        folder_name = os.path.join(
            self.results_folder,
            label,
            frame_label,
            grid_label
        )
        file_name = os.path.join(folder_name, bandwidth_label)
        return file_name, folder_name

    def _add_entry_to_bandwidth_table(self, bandwidth, bandwidth_hash, label):
        table = self._load_bandwidth_table(label)
        if table is None:
            table = self._create_bandwidth_table()
        table.loc[hash(self.grid_generator)] = bandwidth_hash
        file_name = self._construct_bandwidth_table_name(label)
        dir_name = os.path.split(file_name)[0]
        os.makedirs(dir_name, exist_ok=True)
        self._write_bandwidth_to_file(bandwidth, bandwidth_hash, label)
        table['bandwidth_hash'] = table['bandwidth_hash'].astype(int)
        table.to_csv(file_name)

    def _load_bandwidth_table(self, label) -> Union[None, pd.DataFrame, object]:
        file_name = self._construct_bandwidth_table_name(label)
        if not os.path.exists(file_name):
            return None
        table = pd.read_csv(file_name, index_col=0)
        table['bandwidth_hash'] = table['bandwidth_hash'].astype(int)
        return table

    @staticmethod
    def _create_bandwidth_table() -> pd.DataFrame:
        index = pd.Series(name='grid_hash', dtype=int)
        table = pd.DataFrame(columns=['bandwidth_hash'], index=index)
        return table

    def _load_average_densities(
            self,
            label: str,
            frame: int,
            bandwidth: Union[None, float],
            bandwidth_hash: Union[None, int]
    ) -> Tuple[Union[None, np.ndarray],
               Union[None, float],
               Union[None, int]]:
        if bandwidth is None:
            bandwidth, bandwidth_hash = self._load_bandwidth_from_table(label)
        if bandwidth is None:
            return None, bandwidth, bandwidth_hash

        file_name, folder_name = self._construct_folder_and_file_name(bandwidth_hash, frame, label)
        if not os.path.exists(file_name):
            return None, bandwidth, bandwidth_hash
        return np.load(file_name), bandwidth, bandwidth_hash

    def _load_bandwidth_from_table(
            self,
            label: str
    ) -> Union[Tuple[None, None], Tuple[float, int]]:
        table = self._load_bandwidth_table(label)
        if table is None:
            return None, None
        if hash(self.grid_generator) not in table.index:
            return None, None
        bandwidth_hash = table.loc[hash(self.grid_generator)]
        bandwidth = self._load_bandwidth_from_file(label, bandwidth_hash)
        return bandwidth, bandwidth_hash

    @staticmethod
    def _get_kde(
            simulation: Simulation,
            frame: int,
            bandwidth: Union[float, None]
    ) -> gaussian_kde:
        return simulation.analyser.get_densities_gaussian_kde_scipy(frame, bandwidth)

    @staticmethod
    def _get_density_grid_single(
            kde: gaussian_kde,
            X: np.ndarray,
            Y: np.ndarray,
            Z: np.ndarray
    ) -> np.ndarray:
        positions = np.vstack([X.ravel(), Y.ravel(), Z.ravel()])
        dens = kde(positions)
        dens = dens.reshape(X.shape)
        return dens

    def _construct_bandwidth_table_name(self, label) -> str:
        return os.path.join(self.results_folder, label, 'bandwidth_hashes.csv')

    def _write_bandwidth_to_file(
            self,
            bandwidth,
            bandwidth_hash,
            label
    ) -> None:
        folder_name = os.path.join(self.results_folder, label)
        file_name = os.path.join(folder_name, f'bandwidth_{bandwidth_hash}.npy')
        os.makedirs(folder_name, exist_ok=True)
        np.save(file_name, np.array([bandwidth]))

    def _load_bandwidth_from_file(self, label: str, bandwidth_hash: int) -> float:
        folder_name = os.path.join(self.results_folder, label)
        file_name = os.path.join(folder_name, f'bandwidth_{bandwidth_hash}.npy')
        return np.load(file_name)[0]


_DensityResults = GroupResultsContainer[KDEResult]



from typing import Generic, TypeVar, Dict

from actomyosin_analyser.tools.experiment_iterator import GroupIterator

T = TypeVar('T')


class GroupResultsContainer(Generic[T]):
    """
    Container to store results for groups (see GroupIterator class).
    Results can be accessed via ``GroupIterator`` instances, e.g.
    ``value = container[group]``, where ``container`` is an instance
    of ``GroupResultsContainer``.
    """

    def __init__(self) -> None:
        self.items: Dict[str, T] = {}

    def __setitem__(self, group: GroupIterator, value: T):
        key = group.get_label_from_values()
        if key in self.items:
            raise KeyError(f"Results for group with label {key} are already"
                           "in present int he container.")
        self.items[key] = value

    def __getitem__(self, group: GroupIterator) -> T:
        key = group.get_label_from_values()
        if key not in self.items:
            raise KeyError(f"No results for group with label {key} found.")
        return self.items[key]



import numpy as np
import numba

BooleanArray = np.ndarray


def get_bounding_box(coords: np.ndarray) -> np.ndarray:
    bb = np.empty((3, 2))
    bb[:, 0] = coords.min(0)
    bb[:, 1] = coords.max(0)
    return bb


def bounding_boxes_might_be_in_proximity(proximity, bb,
                                         bb_other, min_img_box) -> BooleanArray:
    center = bb.sum(1) / 2
    d = bb[:, 1] - bb[:, 0]
    centers_other = bb_other.sum(2) / 2
    d_other = bb_other[:, :, 1] - bb_other[:, :, 0]

    cvector = np.abs(get_minimum_image_vector(center, centers_other, min_img_box))

    potential_proximity = (cvector - (d_other + d) / 2 - proximity) <= 0

    return potential_proximity.all(1)


@numba.jit
def get_minimum_image_vector(v1, v2, simulation_box):
    """
    Minimum image vector from coordinates v2 to v1.
    Either v1 or v2 needs to contain multiple coordinates, i.e. has
    to have ndim = 2. E.g., v1 is of shape (5, 3) (holds coordinates of
    5 particles), v2 is of shape (3,).
    """
    dv = v1 - v2
    for i in range(len(simulation_box)):
        size_i = simulation_box[i]
        dv[:, i] -= size_i * np.rint(dv[:, i] / size_i)
    return dv



import os
from typing import Callable, Dict, Any, Iterable, List, Tuple, Union, Sequence
import math
import json
import logging

from scipy import stats
import numpy as np
import numba
import h5py
from actomyosin_analyser.file_io.data_reader import DataReader
from . import geometry
from .abstract_analyser import AbstractAnalyser
from .filament_center_of_mass import FilamentCenterOfMassGetter
from .polymer_density import GridGenerator, PolymerDensityGetter
from .mean_squared_displacement import compute_msd, MeanSquaredDisplacement
from .._version import get_version
from ..file_io.state_exporter import StateExporter
from ..model.bead import Filament


_VERSION = get_version()
logger = logging.getLogger(__name__)

Dataset = h5py.Dataset
_FilamentT = Tuple[np.ndarray, np.ndarray]
_FilamentsT = numba.typed.List  # [_FilamentT]


class Analyser(AbstractAnalyser):
    """
    The ``Analyser`` class defines the user interface to the analysis tools
    of the ``actomyosin_analyser`` package. Use the methods of an
    ``Analyser`` instance to run the analysis on your data.
    Reading of the raw data is delegated to the ``DataReader`` provided when
    creating the instance of the ``Analyser``.
    More on the :class:`~actomyosin_analyser.file_io.data_reader.DataReader` class
    can be found in its documentation.
    """

    def __init__(self, data_reader: DataReader, analysis_file: str):
        """
        Create an instance of the ``Analyser`` class. Provide an instance
        of a ``DataReader`` that handles reading the raw data (``DataReader``
        has to be implemented depending on the format of your data), and
        the path to an analysis file (hdf5 format, usual file extension is `.h5`).

        :param data_reader: Instance of a :class:`~actomyosin_analyser.file_io.data_reader.DataReader`
        :param analysis_file: Path to the analysis file
           (usually `"some/folder/analysis.h5"` in the examples).
        """
        _data_root = os.path.split(analysis_file)[0]
        if len(_data_root) == 0:
            _data_root = '.'
        super().__init__(_data_root)
        self.data_reader = data_reader
        self.analysis_file = analysis_file
        self._offset_filaments = None
        self._particle_positions = None
        self._analysis_parameter_file = os.path.join(self.blob_root, 'parameters.json')

    def _get_analysis_dataset(self,
                              dset_name: str,
                              compute_function: Callable[[Any], np.ndarray],
                              **keyword_arguments: Any) -> np.ndarray:
        kwargs_as_json = None
        if keyword_arguments:
            kwargs_as_json = Analyser._keyword_arguments_to_json_string(keyword_arguments)
            dset_name += '_' + kwargs_as_json
        if os.path.exists(self.analysis_file):
            with h5py.File(self.analysis_file, 'r') as fp:
                if dset_name in fp:
                    dataset = fp[dset_name]
                    if keyword_arguments:
                        kwargs = json.loads(dataset.attrs['parameters'])
                        if kwargs != keyword_arguments:
                            msg = "DataSet '{}' has matching name, but loaded parameters are different!"
                            msg = msg.format(dataset.name)
                            raise RuntimeError(msg)
                    return np.array(dataset)
        data = compute_function(**keyword_arguments)
        with h5py.File(self.analysis_file, 'a') as fp:
            fp[dset_name] = data
            fp[dset_name].attrs['ACTOMYOSIN_ANALYSER_VERSION'] = _VERSION
            if kwargs_as_json is not None:
                fp[dset_name].attrs['parameters'] = kwargs_as_json
        return data

    def _get_nested_analysis_dataset(self, group_name: str,
                                     compute_function: Callable[[Any], np.ndarray],
                                     **keyword_arguments: Any) -> np.ndarray:
        if not keyword_arguments:
            msg = "keyword_arguments can not be empty for Analyser._get_nested_analysis_dataset. "
            msg += "If you have no keywords, use Analyser._get_analysis_dataset instead!"
            raise KeyError(msg)
        kwargs_as_json = Analyser._keyword_arguments_to_json_string(keyword_arguments)
        h = kwargs_as_json

        with h5py.File(self.analysis_file, 'a') as fp:
            if group_name in fp:
                group = fp[group_name]
            else:
                group = fp.create_group(group_name)

            if h in group:
                dset = group[h]
                kwargs = json.loads(dset.attrs['parameters'])
                for_comparison = keyword_arguments.copy()
                for k in for_comparison.keys():
                    value = for_comparison[k]
                    if isinstance(value, tuple):
                        for_comparison[k] = 'hashed_tuple_' + str(hash(value))
                if kwargs != for_comparison:
                    msg = "DataSet '{}' has matching name, but loaded parameters are different!".format(
                        dset.name)
                    raise RuntimeError(msg)
                return np.array(dset)

            data = compute_function(**keyword_arguments)
            group[h] = data
            group[h].attrs['parameters'] = kwargs_as_json
            group[h].attrs['ACTOMYOSIN_ANALYSER_VERSION'] = _VERSION
        return data

    def _get_parameter(self, parameter_name,
                       compute_function: Callable[[], Any]) -> Any:
        if os.path.exists(self._analysis_parameter_file):
            with open(self._analysis_parameter_file, 'rt') as fh:
                parameters = json.load(fh)
        else:
            parameters = {}
        if parameter_name in parameters:
            return parameters[parameter_name]
        par = compute_function()
        parameters[parameter_name] = par
        with open(self._analysis_parameter_file, 'wt') as fp:
            json.dump(parameters, fp)
        return par

    def get_offset_filaments(self) -> int:
        return self._get_parameter('offset_filaments',
                                   self._determine_offset_filaments)

    def _determine_offset_filaments(self) -> int:
        if self._offset_filaments is not None:
            return self._offset_filaments
        self._offset_filaments = self.data_reader.get_n_non_filament_particles()
        return self._offset_filaments

    def get_filaments(self, save: bool = False) -> List[List[Filament]]:
        return self._get_filaments_from_analysis_file_or_data_reader(save)

    def get_trajectories_filaments(self, minimum_image: bool) -> np.ndarray:
        """
        Load complete trajectory of all filaments.
        Filaments are represented by multiple coordinates,
        which corresponds to coordinates of beads in bead chains (e.g. from simulations
        with the ``bead_state_model`` framework)
        , or start and end points
        of segments in segment representations of filaments (e.g. from simulations with
        the ``cytosim`` framework).
        Which coordinates belong to which filament can be figured out with the return
        value of the ``get_filaments`` method.

        This can be very
        memory intensive for ``bead_state_model`` data, when executed for the first time,
        depending of the number of particles and recorded frames.
        When this method is executed,  it will save the trajectory to the analysis file,
        which has the benefit of much quicker and way less memory intensive execution
        on the next call. However, this additional data saved to the analysis file
        might double the storage space needed, since it will inflate the analysis file
        to a similar size of the raw data file.
        If you only want use a small subset of frames, and you want to save storage space,
        use the method get_positions_of_filaments_at_frames.

        :param minimum_image: Set ``True`` to project coordinates that exceed
           the simulation box back into the simulation box.
        :return: Numpy array with 3 dimensions:
           (number of frames, number of filament particles, xyz-coordinates).
        """
        return self._get_analysis_dataset('trajectories_filaments',
                                          self._get_trajectories_filament,
                                          minimum_image=minimum_image)

    def get_trajectories_non_filament(self, minimum_image: bool) -> np.ndarray:
        """
        Get trajectories on non-filament particles. These are all particles
        that are not part of filaments. For example passive beads of
        a microrheology experiment/simulation. The result does not contain
        information of the particle species.

        :param minimum_image: Set ``True`` to project coordinates that exceed
           the simulation box back into the simulation box.
        :return: Numpy array with 3 dimensions:
           (number of frames, number of particles, xyz-coordinates).
        """
        return self._get_analysis_dataset('trajectories_non_filament',
                                          self._get_trajectories_non_filament,
                                          minimum_image=minimum_image)

    def get_positions_of_filaments_at_frames(
            self,
            frames: Sequence[int],
            minimum_image: bool = False
    ) -> np.ndarray:
        """
        Same as method ``get_trajectories_filaments``, but for a subset of frames,
        rather than for all frames.
        See documentation of method
        :py:meth:`~actomyosin_analyser.analysis.analyser.Analyser.get_trajectories_filaments`.

        :param frames: Selected subset of frames.
        :param minimum_image: Set ``True`` to project coordinates that exceed
           the simulation box back into the simulation box.
        :return: Numpy array with 3 dimensions:
           (number of selected frames, number of filament particles, xyz-coordinates).
        """
        positions = self._load_positions_of_filaments_at_frames(frames, minimum_image)
        if positions is not None:
            return positions
        positions = self.data_reader.get_filament_coordinates(minimum_image)[frames]
        self._save_particle_positions_at_frames(frames, positions, minimum_image)
        return positions

    def export_state(self, frame: int):
        """
        Export the state of the system at a single frame. The state includes
        information of all filament coordinates, links between filament beads, and
        positions of all external particles.
        State data is written to folder `analysis_blobs/state` placed next to the
        analysis_file (specified during creation of the ``Analyser`` instance).

        :param frame: Selected frame.
        """
        state_exporter = StateExporter(self)
        state_exporter.export(frame)

    def _load_positions_of_filaments_at_frames(
            self,
            frames,
            minimum_image
    ) -> Union[None, np.ndarray]:
        if self._contains_data_set('trajectories_filaments', minimum_image=minimum_image):
            return self.get_trajectories_filaments(minimum_image)[frames]
        positions = []
        for f in frames:
            p = self._load_positions_of_filaments_at_single_frame(f, minimum_image)
            if p is None:
                return None
            positions.append(p[np.newaxis])
        return np.concatenate(positions)

    def _contains_data_set(self, group_label: str, **kwargs) -> bool:
        if not os.path.exists(self.analysis_file):
            return False
        with h5py.File(self.analysis_file, 'r') as h5f:
            if kwargs:
                kwargs_as_json = Analyser._keyword_arguments_to_json_string(kwargs)
                group_label += kwargs_as_json
            return group_label in h5f

    def _save_particle_positions_at_frames(
            self,
            frames: Sequence[int],
            positions: np.ndarray,
            minimum_image: bool
    ):
        for i, f in enumerate(frames):
            p = positions[i]
            self._save_particle_positions_at_single_frame(f, p, minimum_image)

    def _save_particle_positions_at_single_frame(
            self,
            frame: int,
            positions: np.ndarray,
            minimum_image: bool
    ):
        label = "filament_positions_at_frame"
        if minimum_image:
            label = 'minimum_image_' + label
        with h5py.File(self.analysis_file, 'a') as h5f:
            if label in h5f:
                group = h5f[label]
            else:
                group = h5f.create_group(label)
            group[str(frame)] = positions

    def _load_positions_of_filaments_at_single_frame(
            self,
            frame: int,
            minimum_image: bool
    ) -> Union[np.ndarray, None]:
        if not os.path.exists(self.analysis_file):
            return None
        label = "filament_positions_at_frame"
        if minimum_image:
            label = 'minimum_image_' + label
        with h5py.File(self.analysis_file, 'r') as h5f:
            if label not in h5f:
                return None
            group = h5f[label]
            if str(frame) not in group:
                return None
            return group[str(frame)][:]

    def get_filament_center_of_mass_msds(self, skip: int) -> MeanSquaredDisplacement:
        """
        Get mean-squared displacements (MSDs) of the centers of mass of all filaments.
        Resulting ensemble MSD is average over individual filament center of mass MSDs.

        :param skip: Initial frames to be skipped, in case the state is not in the desired
           state from the start. Set to 0 if no frames are to be skipped.
        :return: ``MeanSquaredDisplacement`` instance; a dataclass with 4 attributes:
           ``lag`` (in frames), ``lag_time``, ``msd``, ``variance_of_msd``.
        """
        dataset_name = 'filament_center_of_mass_msds'
        keyword_args = {'skip': skip}
        msd_array = self._get_nested_blob_analysis_dataset(
            dataset_name,
            self._compute_filament_center_of_mass_msds,
            keyword_args
        )
        msd = MeanSquaredDisplacement.from_array(msd_array)
        time = self.get_time()
        msd.lag_time = time[msd.lag.astype(int)]
        return msd

    def _compute_filament_center_of_mass_msds(self, skip: int) -> np.ndarray:
        fcom_getter = FilamentCenterOfMassGetter(self)
        fcom_trajectories = fcom_getter.get()
        fcom_traj_array = fcom_trajectories.as_array()[skip:]
        box = self.read_simulation_box()
        msd = compute_msd(fcom_traj_array, box)
        return msd.as_array()

    def get_msd_ensemble(self, particle_indices: List[int], skip=0) -> np.ndarray:
        return self._get_nested_analysis_dataset('msd_ensemble', self._compute_msd_ensemble,
                                                 particle_indices=tuple(particle_indices),
                                                 skip=skip)

    def get_msd(self, particle_indices: List[int], non_filament: bool = False, skip=0) -> List[MeanSquaredDisplacement]:
        if non_filament:
            method = self.get_msd_single_particle_non_filament
        else:
            method = self.get_msd_single_particle
        msds = []
        for i in range(len(particle_indices)):
            msds.append(method(particle_indices[i], skip))
        return msds

    def get_msd_single_particle(self, particle_index: int, skip: int) -> MeanSquaredDisplacement:
        msd_array = self._get_nested_analysis_dataset('msd_filament', self._compute_msd,
                                                      particle_index=particle_index,
                                                      skip=skip)
        return MeanSquaredDisplacement.from_array(msd_array)

    def get_msd_single_particle_non_filament(self, particle_index: int, skip: int) -> MeanSquaredDisplacement:
        msd_array = self._get_nested_analysis_dataset('msd_non_filament', self._compute_msd_non_filament,
                                                      particle_index=particle_index,
                                                      skip=skip)
        return MeanSquaredDisplacement.from_array(msd_array)

    def get_polymer_density(
            self,
            grid: GridGenerator,
            frames: Sequence[int],
            bandwidth: Union[float, None] = None
    ) -> Tuple[float, np.ndarray]:
        """
        Get the polymer density in the simulation box. Density is estimated
        at points on a grid. Estimation is done via scipy's gaussian_kde function.
        If no bandwidth is specified, bandwidth is generated automatically during
        evaluation of the density at the first frame. Bandwidth has to
        be identical across frames, in order for the densities to be comparable
        across frames.

        :param grid: Instance of :class:`~actomyosin_analyser.analysis.polymer_density.GridGenerator`
           specifying the grid at which the density gets evaluated.
        :param frames: Selected frames.
        :param bandwidth: Bandwidth used for Gaussian kernel density estimation (see scipy documentation).
        :return: Tuple containing (1) the used bandwidth and
           (2) the density evaluated at grid points.
        """
        density_getter = PolymerDensityGetter(grid, self)
        bandwidth, density = density_getter.get(frames, bandwidth)
        return bandwidth, density

    def get_densities_gaussian_kde_scipy(
            self, frame: int,
            bandwidth: Union[float, None] = None
    ) -> stats.kde.gaussian_kde:
        positions = self.get_positions_of_filaments_at_frames([frame], minimum_image=True)[0]
        if bandwidth is None:
            kde = stats.gaussian_kde(positions.T)
        else:
            kde = stats.gaussian_kde(positions.T, bw_method=bandwidth)
        return kde

    def get_links(self, frame: int) -> np.ndarray:
        return self.data_reader.get_filaments(frame)[1]

    def get_link_count(self) -> np.ndarray:
        return self._get_blob_analysis_dataset('link_count', self.data_reader.get_motor_count)

    def get_time(self) -> np.ndarray:
        return self._get_blob_analysis_dataset('time', self.data_reader.read_time)

    def get_contour_lengths(self) -> np.ndarray:
        return self._get_analysis_dataset('contour_lengths',
                                          self._compute_contour_lengths)

    def get_end_to_end_distances(self, t_frame: int, n_beads: int) -> np.ndarray:
        """
        Get end-to-end distances of filaments with n_beads at frame t_frame.
        """
        return self._get_nested_blob_analysis_dataset('end_to_end_distances',
                                                      self._compute_end_to_end_distances,
                                                      {'t': t_frame, 'n_beads': n_beads})

    def get_end_to_end_vectors(self) -> np.ndarray:
        """
        Get end-to-end vectors for all filaments at all frames. 
        
        :return: Array with 3 dimensions: (1) time frame (2) filament ID (3) XYZ coordinates.
        """
        return self._get_analysis_dataset('end_to_end_vectors', self._get_end_to_end_vectors)

    def get_time_in_steps(self) -> np.ndarray:
        return self._get_analysis_dataset('time_in_steps', self._get_time_in_steps)

    def get_energies_stretch_and_bend(self) -> np.ndarray:
        """
        For data of a ``bead_state_model`` simulation, get bending and stretching energy of filaments.
        For other sources of data, you need to implement your own method to compute energies.

        :return: Numpy array with 2 dimensions:
           (number of frames, energy (2 separate columns for stretch and bend))
        """
        energies = self._get_analysis_dataset('energies', self._compute_energies_stretch_and_bend)
        return energies

    def _get_time_in_steps(self):
        return self.data_reader.read_time_as_step_indices()

    def _get_trajectories_non_filament(self, minimum_image: bool) -> np.ndarray:
        pos = self.data_reader.get_non_filament_coordinates(minimum_image)
        return pos

    def _get_trajectories_filament(self, minimum_image: bool):
        pos = self.data_reader.read_particle_positions(minimum_image)
        offset = self.get_offset_filaments()
        return pos[:, offset:]

    def _convert_binning_parameters_to_lists(self, n_bins, ranges):
        if ranges is None:
            ranges = self.data_reader.read_box_size()
        if isinstance(ranges, np.ndarray):
            range_x = ranges[0].tolist()
            range_y = ranges[1].tolist()
            range_z = ranges[2].tolist()
        else:
            range_x = list(ranges[0])
            range_y = list(ranges[1])
            range_z = list(ranges[2])
        if isinstance(n_bins, np.ndarray):
            n_bins = n_bins.tolist()
        n_bins = list(n_bins)
        return n_bins, range_x, range_y, range_z

    def _compute_msd(self, particle_index: int, skip: int) -> np.ndarray:
        coords = self.get_trajectories_filaments(minimum_image=False)[skip:, particle_index]
        return compute_msd(coords).as_array()

    def _compute_msd_non_filament(self, particle_index: int, skip: int) -> np.ndarray:
        traj = self.get_trajectories_non_filament(minimum_image=False)[skip:, particle_index]
        return compute_msd(traj).as_array()

    def _compute_msd_ensemble(self, particle_indices: Tuple[int], skip: int):
        coords = self.get_trajectories_filaments()[skip:, particle_indices]
        return compute_msd(coords).as_array()

    def _compute_densities(self,
                           particle_types: List[str],
                           range_x: List[float],
                           range_y: List[float],
                           range_z: List[float],
                           n_bins: List[int],
                           count: bool) -> np.ndarray:
        wx = (range_x[1] - range_x[0]) / n_bins[0]
        wy = (range_y[1] - range_y[0]) / n_bins[1]
        wz = (range_z[1] - range_z[0]) / n_bins[2]
        V = wx * wy * wz

        positions = self.data_reader.read_particle_positions(minimum_image=True)
        pmap, ptypes = self.data_reader.read_particle_types()

        shape = (len(positions), n_bins[0], n_bins[1], n_bins[2])
        densities = np.full(shape, np.nan)

        for t in range(positions.shape[0]):
            pos_t = positions[t]
            ptypes_t = ptypes[t]
            mask = np.zeros_like(ptypes_t, dtype=bool)
            for pt in particle_types:
                mask += (ptypes_t == pmap[pt])
            selected_particles = pos_t[mask]
            bin_indices = Analyser._coordinates_to_bin_indices(selected_particles,
                                                               range_x, wx,
                                                               range_y, wy,
                                                               range_z, wz)
            for i in range(shape[1]):
                for j in range(shape[2]):
                    for k in range(shape[3]):
                        count_ijk = ((bin_indices[:, 0] == i) &
                                     (bin_indices[:, 1] == j) &
                                     (bin_indices[:, 2] == k)).sum()
                        densities[t, i, j, k] = count_ijk
        if count:
            densities = densities.astype(int)
        else:
            densities = densities / V

        return densities

    def _compute_end_to_end_distances(self, t: int, n_beads: int) -> np.ndarray:
        coordinates = self.get_trajectories_filaments(minimum_image=True)[t]
        filaments, _ = self.data_reader.get_filaments(t)
        box = self.read_simulation_box()

        e2e = []
        for fil in filaments:
            if len(fil.items) != n_beads:
                continue
            e2e.append(self._compute_e2e_single_filament(coordinates, fil, box))
        return np.array(e2e)

    def _get_end_to_end_vectors(self) -> np.ndarray:
        raise NotImplementedError
        coordinates = self.get_trajectories_filaments()
        filaments = self.data_reader.get_filaments_all()

    def _compute_e2e_single_filament(self, coordinates: np.ndarray, filament, box) -> float:
        offset_filaments = self.get_offset_filaments()
        fcoords = coordinates[np.array(filament.items) + offset_filaments]
        v_segments = geometry.get_minimum_image_vector(fcoords[:-1], fcoords[1:], box)
        e2e = np.sum(v_segments, axis=0)
        return math.sqrt(e2e[0] ** 2 + e2e[1] ** 2 + e2e[2] ** 2)

    def _compute_contour_lengths(self) -> np.ndarray:
        coordinates = self.get_trajectories_filaments()
        box = self.read_simulation_box()

        filaments = self.get_filaments()

        contour_lengths = np.full((len(coordinates), len(filaments[0])), np.nan)

        for t in range(len(coordinates)):
            fil_t = filaments[t]
            for f, fil in enumerate(fil_t):
                fcoords = coordinates[t, fil.items]
                v_segments = geometry.get_minimum_image_vector(fcoords[:-1], fcoords[1:], box)
                d = np.sqrt(np.sum(v_segments ** 2, 1))
                lc = np.sum(d)
                contour_lengths[t, f] = lc

        return contour_lengths

    def read_simulation_box(self):
        box = self.data_reader.read_box_size()
        box = np.array([box[1, 0] - box[0, 0],
                        box[1, 1] - box[0, 1],
                        box[1, 2] - box[0, 2]])
        return box

    def _compute_energies_stretch_and_bend(self) -> np.ndarray:
        coordinates = self.get_trajectories_filaments()
        box = self.read_simulation_box()

        filaments = self.get_filaments()
        filaments_as_tuples = Analyser._convert_nested_filaments_to_tuples(filaments)

        return Analyser._numba_compute_energies_stretch_and_bend(box, coordinates,
                                                                 filaments_as_tuples)

    @staticmethod
    def _convert_nested_filaments_to_tuples(filaments) -> numba.typed.List:
        filaments_as_tuples = numba.typed.List()
        for f_frame in filaments:
            filaments_as_tuples_single_frame = numba.typed.List()
            for f in f_frame:
                filaments_as_tuples_single_frame.append((
                    np.array(f.items).astype('uint32'),
                    np.array(f.motors).astype('uint32')))
            filaments_as_tuples.append(filaments_as_tuples_single_frame)
        return filaments_as_tuples

    @staticmethod
    @numba.njit
    def _numba_compute_energies_stretch_and_bend(
            box, coordinates,
            filaments: numba.typed.List  # List[_FilamentsT]
    ):
        energies = np.zeros((len(coordinates), 2))
        for t in range(len(coordinates)):
            fil_t = filaments[t]
            energies_t = np.zeros((len(fil_t), 2))
            for f, fil in enumerate(fil_t):
                coords_t = coordinates[t]
                fcoords = coords_t[fil[0]]
                v_segments = geometry.get_minimum_image_vector(
                    fcoords[:-1], fcoords[1:], box)
                segment_lengths = np.sqrt(np.sum(v_segments ** 2, 1))

                e_stretch = np.sum((segment_lengths - 1) ** 2)

                cos_theta = ((v_segments[1:] * v_segments[:-1]).sum(1)
                             / segment_lengths[1:] / segment_lengths[:-1])

                e_bend = np.sum(np.arccos(cos_theta) ** 2)

                energies_t[f, 0] = e_stretch
                energies_t[f, 1] = e_bend

            energies[t] = np.sum(energies_t, axis=0)
        return energies

    @staticmethod
    def _keyword_arguments_to_json_string(keyword_arguments: Dict[str, Any]) -> str:
        """
        Note that tuples get hashed. This is to avoid containers with thousands of values
        to become part of the json string, which will be used as key to retrieve
        data from file. Make sure you provide containers with countless items as tuples.
        """
        if not keyword_arguments:
            return ''
        sorted_dict = {}
        for key in sorted(keyword_arguments.keys()):
            value = keyword_arguments[key]
            if isinstance(value, tuple):
                value = 'hashed_tuple_' + str(hash(value))
            sorted_dict[key] = value
        return json.dumps(sorted_dict)

    @staticmethod
    def _get_voxel_mask(bin_indices, grid_range, indices, shape) -> np.ndarray:
        i, j, k = indices
        mask_i = ((np.abs((bin_indices[:, 0]) - i) <= grid_range[0]) |
                  ((-np.abs(bin_indices[:, 0] - i) % shape[0]) <= grid_range[0]))
        mask_j = ((np.abs((bin_indices[:, 1]) - j) <= grid_range[1]) |
                  ((-np.abs(bin_indices[:, 1] - j) % shape[1]) <= grid_range[1]))
        mask_k = ((np.abs((bin_indices[:, 2]) - k) <= grid_range[2]) |
                  ((-np.abs(bin_indices[:, 2] - k) % shape[2]) <= grid_range[2]))
        mask_ijk = (
                mask_i & mask_j & mask_k
        )
        return mask_ijk

    @staticmethod
    def _coordinates_to_bin_indices(coordinates: np.ndarray,
                                    range_x: List[float], width_x: float,
                                    range_y: List[float], width_y: float,
                                    range_z: List[float], width_z: float) -> np.ndarray:
        bin_indices = np.full((len(coordinates), 3), -1, dtype=int)
        for i, (range_i, width) in enumerate(zip([range_x, range_y, range_z],
                                                 [width_x, width_y, width_z])):
            mask = (coordinates[:, i] >= range_i[0]) & (coordinates[:, i] <= range_i[1])
            bin_indices[mask, i] = (coordinates[mask, i] - range_i[0]) // width
        return bin_indices

    def _get_filaments_from_analysis_file_or_data_reader(self, save: bool) -> List[List[Filament]]:
        if os.path.exists(self.analysis_file):
            with h5py.File(self.analysis_file, 'r') as fp:
                if 'filaments' in fp:
                    n_frames = len(self.get_time())
                    return Analyser._load_filaments_from_h5_group(fp['filaments'],
                                                                  n_frames)
        filaments = self.data_reader.get_filaments_all()
        if save:
            with h5py.File(self.analysis_file, 'a') as fp:
                group = fp.create_group('filaments')
                Analyser._write_filaments_to_h5_group(group, filaments)
        return filaments

    @staticmethod
    def _load_filaments_from_h5_group(group: h5py.Group, n_frames: int) -> List[List[Filament]]:
        filaments = []
        for f in range(n_frames):
            filaments_f = Analyser._load_filaments_from_h5_group_single_frame(group[str(f)])
            filaments.append(filaments_f)
        return filaments

    @staticmethod
    def _load_filaments_from_h5_group_single_frame(group: h5py.Group) -> List[Filament]:
        items = group['items'][:]
        motors = group['motors'][:]
        filament_tuples = Analyser._parse_items_and_motors_arrays_to_filament_tuples(items, motors)
        filaments = [Filament(list(ftuple[0]), list(ftuple[1])) for ftuple in filament_tuples]
        return filaments

    @staticmethod
    @numba.njit
    def _parse_items_and_motors_arrays_to_filament_tuples(
            items: np.ndarray, motors: np.ndarray
    ) -> _FilamentsT:
        tuples = numba.typed.List()
        assert len(items) == len(motors)
        _invalid = np.iinfo(items.dtype).max
        for i in range(len(items)):
            items_i = items[i]
            items_i = items_i[items_i != _invalid]
            motors_i = motors[i]
            motors_i = motors_i[motors_i != _invalid]
            tuples.append((items_i, motors_i))
        return tuples

    @staticmethod
    def _write_filaments_to_h5_group(group: h5py.Group, filaments: List[List[Filament]]):
        for frame in range(len(filaments)):
            filaments_t = filaments[frame]
            typed_list = numba.typed.List()
            [typed_list.append((np.array(f.items).astype('uint32'),
                                np.array(f.motors).astype('uint32'))) for f in filaments_t]
            group_t = group.create_group(str(frame))
            Analyser._write_filaments_to_h5_group_single_frame(group_t, typed_list)

    @staticmethod
    def _write_filaments_to_h5_group_single_frame(
            group: h5py.Group, filaments: _FilamentsT):
        max_idx, max_items, max_motors = Analyser._get_maximum_items_and_value_for_filaments(filaments)

        items, motors = Analyser._create_filament_item_and_motor_arrays(len(filaments),
                                                                        max_idx,
                                                                        max_items,
                                                                        max_motors)

        Analyser._fill_filament_item_and_motor_arrays(filaments, items, motors)

        group.create_dataset('items', data=items, compression='lzf')
        group.create_dataset('motors', data=motors, compression='lzf')

    @staticmethod
    @numba.njit
    def _fill_filament_item_and_motor_arrays(
            filaments: _FilamentsT,
            items: np.ndarray, motors: np.ndarray):
        for i, (items_i, motors_i) in enumerate(filaments):
            items[i, :len(items_i)] = items_i
            motors[i, :len(motors_i)] = motors_i

    @staticmethod
    @numba.njit
    def _get_maximum_items_and_value_for_filaments(
            filaments: _FilamentsT
    ) -> Tuple[int, int, int]:
        max_items = 0
        max_idx = 0
        max_motors = 0
        for items_i, motors_i in filaments:
            max_idx_f = max(items_i)
            if max_idx_f > max_idx:
                max_idx = max_idx_f
            if len(items_i) > max_items:
                max_items = len(items_i)
            if len(motors_i) > max_motors:
                max_motors = len(motors_i)
        return max_idx, max_items, max_motors

    @staticmethod
    def _create_filament_item_and_motor_arrays(
            n_filaments: int, max_idx: int,
            max_items: int, max_motors: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        dtype = 'uint32'
        _invalid = np.iinfo(dtype).max
        if max_idx > _invalid:
            raise ValueError("can't handle indices larger than "
                             f"{_invalid}, which is the maximum "
                             "for 32 bit unsigned integers")

        items = np.full((n_filaments, max_items), _invalid, dtype=dtype)
        motors = np.full((n_filaments, max_motors), _invalid, dtype=dtype)
        return items, motors



import numpy as np
from mayavi.mlab import points3d, draw


class FilamentAnimator:

    def __init__(self):
        self.filament_data_reader = None
        self.points = None

    def show_at_t(self, t, figure=None):
        self._set_render_disable_status(True)
        data_slice_t, _ = self.filament_data_reader.get_coordinates_and_status_at_time_step(t)
        self._draw_data_slice(data_slice_t, figure)
        # self._set_render_disable_status(False)

    def _set_render_disable_status(self, state: bool):
        self.points.scene.disable_render = state

    def initialize_animation(self, time_step):
        data_slice, _ = self.filament_data_reader.get_coordinates_and_status_at_time_step(time_step)
        self._draw_data_slice(data_slice)

    def _draw_data_slice(self, data_slice, figure=None):
        if data_slice.shape[0] % 3 != 0:
            msg = "data_slice needs to be multiple of 3 (x-, y-, z-coordinates)."
            raise RuntimeError(msg)
        x, y, z = self._data_slice_to_coords(data_slice)
        if figure is not None:
            self.points = points3d(x, y, z,
                                   figure=figure,
                                   scale_mode="none", scale_factor=50.0, resolution=25)
            return
        
        if self.points is None:
            self.points = points3d(x, y, z,
                                   scale_mode="none", scale_factor=50.0, resolution=25)
        else:
            self.points.mlab_source.reset(x=x, y=y, z=z)

    def _data_slice_to_coords(self, data_slice: np.ndarray):
        size = data_slice.shape[0] // 3
        x = np.empty(size)
        y = np.empty(size)
        z = np.empty(size)        

        for i in range(size):
            x[i] = data_slice[3 * i]
            y[i] = data_slice[3 * i + 1]
            z[i] = data_slice[3 * i + 2]
        return x, y, z
